

from fdi.pns.pnsconfig import pnsconfig as pc
from fdi.utils.getconfig import getConfig
# default configuration is provided. Copy contents of svom/engisim/config.py to ~/.config/engisimlocal.py
pc.update(getConfig())

defaulturl = 'http://' + pc['node']['host'] + ':' + \
    str(pc['node']['port']) + pc['baseurl']


def urn2url(urn, contents='product'):
    """ Returns URL for accessing pools with a URN.

    contents:
    'product' for returning a product from the pool.
    'hk' for returning the housekeeping data of the pool.
    'classes' for returning the class housekeeping data of the pool.
    'urns' for returning the URN housekeeping data of the pool.
    'tags' for returning the tag housekeeping data of the pool.

    Example:
    IP=ip dir=/a/b/c files=/a/b//c/classes.jsn | urns.jsn | t.. | urn...

    with python:
    m.refs['myinput'] = special_ref
    ref=pstore.save(m)
    assert ref.urn == 'urn:http://ip:port/a/b/c/fdi.dataset.MapContext:203'
    p=ref.product
    myref=p.refs['myinput']

    with a pool:
    myref=pool.load('http://ip:port/v0.6/a/b/c/fdi.dataset.MapContext/203/refs/myinput')

    urn:http://ip:port/a/b/c/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/a/b/c/fdi.dataset.Product/203/meta/OBSID

    At the same time this is not allowed due to overlapping after 'c'

    urn:http://ip:port/a/b/d/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/a/b/c/d/

    This is also overlapping starting from '/'

    urn:http://ip:port/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/

    but for example this is allowed:

    urn:http://ip:port/a/k/fdi.dataset.Product:203    ==>
    http://ip:port/v0.6/a/k/


    """
    poolname, resourceclass, serialnumstr, scheme, \
        place, poolpath = parseUrn(urn)
    base = poolname + pc['baseurl']
    if contents == 'product':
        ret = base+'/'+resourceclass+'/'+serialnumstr
    elif contents == 'housekeeping':
        ret = base+'/hk'
    elif contents in ['classes', 'urns', 'tags']:
        ret = base+'/hk/' + contents
    else:
        raise(ValueError(contents))

    return ret



# -*- coding: utf-8 -*-

from .common import lls, trbk
import pwd
from os import chown, chmod, environ, setuid, setgid
from subprocess import Popen, PIPE, TimeoutExpired, run as srun

import logging
# create logger
logger = logging.getLogger(__name__)


def run_proc(cmd, as_user, pwdir, timeout):
    """ Execute a shell command and return status.

    """

    try:
        # https://stackoverflow.com/a/6037494
        pw_record = pwd.getpwnam(as_user)
        user_name = pw_record.pw_name
        user_home_dir = pw_record.pw_dir
        user_uid = pw_record.pw_uid
        user_gid = pw_record.pw_gid
        env = environ.copy()
        env['HOME'] = user_home_dir
        env['LOGNAME'] = user_name
        env['PWD'] = pwdir
        env['USER'] = user_name

        def chusr(user_uid, user_gid):
            def result():
                setgid(user_gid)
                setuid(user_uid)
                logger.debug('set uid=%d gid=%d' % (user_uid, user_gid))
            return result

        executable = None

        # /etc/sudoer: apache ALL:(vvpp) NOPASSWD: ALL
        # gpasswd -a vvpp apache
        # cmd = ['sudo', '-u', as_user, 'bash', '-l', '-c'] + cmd
        # cmd = ['sudo', '-u', as_user] + cmd
        logger.debug('Popen %s env:%s uid: %d gid:%d' %
                     (str(cmd), lls(env, 40), user_uid, user_gid))
        proc = Popen(cmd, executable=executable,
                     stdin=PIPE, stdout=PIPE, stderr=PIPE,
                     preexec_fn=None,
                     cwd=pwdir,
                     env=env, shell=False,
                     encoding='utf-8')  # , universal_newlines=True)
    except Exception as e:
        msg = trbk(e)
        return {'returncode': -1, 'message': msg}

    sta = {'command': str(cmd)}

    try:
        sta['stdout'], sta['stderr'] = proc.communicate(timeout=timeout)
        sta['returncode'] = proc.returncode
    except TimeoutExpired:
        # The child process is not killed if the timeout expista,
        # so in order to cleanup properly a well-behaved application
        # should kill the child process and finish communication
        # https://docs.python.org/3.6/library/subprocess.html?highlight=subprocess#subprocess.Popen.communicate
        proc.kill()
        sta['stdout'], sta['stderr'] = proc.communicate()
        sta['returncode'] = proc.returncode

    return sta



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Typed(object):
    """ An object of specifiable kinds. """

    def __init__(self, typ_='', **kwds):
        """ Has a type.

        """

        self.setType(typ_)
        #print(__name__ + str(kwds))
        super(Typed, self).__init__(**kwds)

    @property
    def type(self):
        """ for property getter
        """
        return self.getType()

    @type.setter
    def type(self, typ_):
        """ Must be in ParameterTypes.
        """
        self.setType(typ_)

    def getType(self):
        """ Returns the actual type that is allowed for the value
        of this Parameter."""
        return self._type

    def setType(self, typ_):
        """ Replaces the current type of this parameter.

        """
        self._type = typ_



# -*- coding: utf-8 -*-

import ruamel.yaml
from ruamel.yaml import YAML
from ruamel.yaml.representer import RoundTripRepresenter
from ruamel.yaml.comments import CommentedMap
import inspect
from collections import OrderedDict

from ruamel.yaml.compat import StringIO, ordereddict

import logging
# create logger
logger = logging.getLogger(__name__)


class MyYAML(YAML):
    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


class MyRepresenter(RoundTripRepresenter):
    pass


ruamel.yaml.add_representer(OrderedDict, MyRepresenter.represent_dict,
                            representer=MyRepresenter)

yaml = MyYAML(typ='rt')
yaml.Representer = MyRepresenter
yaml.default_flow_style = False
yaml.width = 80
yaml.allow_unicode = True
yaml.compact(seq_seq=1, seq_map=True)

notinited = True


def yinit(mapping=6, sequence=4, offset=2, register=None):
    """ Initializes YAML.
    """
    global notinited

    yaml.indent(mapping=mapping, sequence=sequence, offset=offset)
    from ..dataset.classes import Classes

    for n, c in Classes.mapping.items():
        if inspect.isclass(c):
            yaml.register_class(c)

    if register:
        for c in register:
            if inspect.isclass(c):
                yaml.register_class(c)

    ruamel.yaml.add_representer(Classes.mapping['ODict'],
                                MyRepresenter.represent_dict,
                                representer=MyRepresenter)
    notinited = False
    return yaml


def ydump(od, stream=None, register=None, **kwds):
    """ YAML dump that outputs OrderedDict like dict.
    """

    global notinited
    if register is None:
        register = []
    if notinited:
        yinit(register=register)

    d = od

    if 0:
        return yaml.dump(d, default_flow_style=False, indent=4,
                         width=60, allow_unicode=True)
    else:
        return yaml.dump(d, stream, **kwds)

# https://stackoverflow.com/a/49048250


if 0:
    def represent_dictionary_order(self, dict_data):
        return self.represent_mapping('tag:yaml.org,2002:map', dict_data.items())

    def setup_yaml():
        yaml.add_representer(OrderedDict, represent_dictionary_order)

    setup_yaml()



# -*- coding: utf-8 -*-
from collections import OrderedDict, UserDict
from collections.abc import Collection
from .serializable import Serializable
from .eq import DeepEqual
from ..utils.common import bstr
from ..utils.ydump import ydump

from pprint import pformat
import logging
import pdb

logger = logging.getLogger(__name__)

# Depth of nesting of ODict.toString()
OD_toString_Nest = 0


class ODict(UserDict, Serializable, DeepEqual):
    """ Ordered dict that is not a subclass of dict and with a better __repr__.
    """

    def __init__(self, *args, **kwds):
        """

        """
        # print(args)
        # data = OrderedDict(*args, **kwds)
        super().__init__(*args, **kwds)
        # UserDict.__init__(self, data)
        Serializable.__init__(self)

    # @property
    # def listoflists(self):
    #     return self.getListoflists()

    # @listoflists.setter
    # def listoflists(self, value):
    #     self.setListoflists(value)

    # def getListoflists(self):
    #     """ Returns a list of lists of key-value pairs where if the key is a tuple or frozenset, it is converted to a list.
    #     """
    #     ret = []
    #     for k, v in self.items():
    #         if issubclass(k.__class__, str):
    #             kk = k
    #         else:
    #             kk = list(k) if issubclass(k.__class__, (Collection)) else k
    #         ret.append([kk, v])
    #     return ret

    # def setListoflists(self, value):
    #     """ Sets the listoflists of this object. """
    #     def c2t(c):
    #         print(c)

    #         lst = [c2t(x) if issubclass(x.__class__, list) else x for x in c]
    #         print('== ', lst)
    #         return tuple(lst)
    #     d = dict(c2t(x) for x in value)
    #     self.clear()
    #     self.update(d)
    #     if 0:
    #         for item in value:
    #             kk = tuple(item[0])
    #             self[kk] = item[1]

    def toString(self, level=0, matprint=None, trans=True, **kwds):
        global OD_toString_Nest

        # return 'OD' + str(type(self.data))+'*'+str(self.data)
        # return 'OD' + str(self.data)
        # return ydump(self.data)

        OD_toString_Nest += 1
        d = ''
        for n, v in self.data.items():
            d += '\n# ' + '    ' * OD_toString_Nest + '[ ' + str(n) + ' ]\n'
            s = bstr(v, level=level, matprint=matprint, trans=trans, **kwds)
            d += s
        OD_toString_Nest -= 1
        return d

    def __repr__(self):
        """ returns string representation with details set according to debuglevel.
        """
        # return 'OD'+super().__repr__()
        level = int(logger.getEffectiveLevel()/10) - 1
        return self.toString(level=level)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(
            data=self.data,
            _STID=self._STID
        )

    def serializable(self):
        """ Can be encoded with serializableEncoder """

        return OrderedDict(
            **self.data,
            _STID=self._STID
        )

    def __hash__(self):

        return hash(tuple(self.data.items()))



# -*- coding: utf-8 -*-
from copy import deepcopy


class Copyable(object):
    """ Interface for objects that can make a copy of themselves. """

    def copy(self):
        """ Makes a deep copy of itself. """
        return deepcopy(self)



# -*- coding: utf-8 -*-
import getopt
import sys

import logging
logger = logging.getLogger(__name__)
# logger.debug('level %d' % (logger.getEffectiveLevel()))


def opt(ops, argv=None):
    """
    Example:
    ops = [
    {'long':'help', 'char':'h', 'default': false, 'description':'print help'},
    {'long':'verbose', 'char':'v', 'default': false, 'description':'print info'},
    {'long':'username=', 'char':'u', 'default': 'foo', 'description':'non-empty user name/ID'},
    {'long':'password=', 'char':'p', 'default': 'bar', 'description':'password'},
    {'long':'host=', 'char':'i', 'default': '0.0.0.0', 'description':'host IP/name'},
    {'long':'port=', 'char':'o', 'default': 5000, 'description':'port number'}
    ]

    Optionly use 'OPTSTART' in command line to start options an have ppreceeding ones ignored. Useful when application is invoked with other options, e.g. by pytest

    argv: if provided will be used in plsce of sys.argv.
    """

    if argv is None:
        argv = sys.argv
    #logger.debug('Input: %s' % ops)
    #stoken = 'OPTSTART'
    #sidx = argv.index(stoken) if stoken in argv else len(argv)

    # this will skip the executable's name at 0
    sidx = 0

    msg = 'Specify:\n'+''.join('%s (-%s or --%s) Default=%s\n\n' %
                               (i['description'], i['char'], i['long'], i['default']) for i in ops)
    # "hu:p:i:o:v"
    fmt = ''.join((i['char']+':' if i['long'].endswith('=')
                   else i['char'] for i in ops))
    try:
        opts, args = getopt.getopt(argv[sidx+1:], fmt, [
                                   i['long'] for i in ops])
    except getopt.GetoptError as err:
        # print help information and exit:
        # will print something like "option -a not recognized"
        logger.error(str(err) + 'Received: ' + str(argv))
        print(msg)
        sys.exit(2)
    logger.debug('Command line options %s args %s' % (opts, args))

    for i in ops:
        switches = ('-'+i['char'], '--'+i['long'].replace('=', ''))
        i['result'] = i['default']
        for o, a in opts:
            if o in switches:
                if i['long'].endswith('='):
                    i['result'] = (i['default'].__class__)(a.lstrip().strip())
                else:
                    i['result'] = True
                    if i['long'] == 'help':
                        print(msg)
                        sys.exit(0)
    # logger.debug(str(ops))
    return ops



# -*- coding: utf-8 -*-

import pdb


class Annotatable(object):
    """ An Annotatable object is an object that can give a
    human readable description of itself.
    """

    def __init__(self, description='UNKNOWN', **kwds):

        self.description = description
        # print(__name__ + str(kwds))
        super(Annotatable, self).__init__(**kwds)

    @property
    def description(self):
        """ xx must be a property for ``self.xx = yy`` to work in super class after xx is set as a property also by a subclass.
        """
        return self.getDescription()

    @description.setter
    def description(self, description):
        self.setDescription(description)

    def getDescription(self):
        """ gets the description of this Annotatable object. """
        return self._description

    def setDescription(self, newDescription):
        """ sets the description of this Annotatable object. """
        self._description = newDescription
        return



# -*- coding: utf-8 -*-

from fdi.dataset.odict import ODict
from fdi.dataset.dataset import TableDataset, IndexedTableDataset
from fdi.dataset.indexed import Indexed
import random
from operator import itemgetter
import gc
import timeit

import string
from timethese import cmpthese, pprint_cmp, timethese

loop = 5
rpt = 2
N = 100000


gc.collect()
gc.disable()

t = [random.randrange(N) for i in range(N)]

print('1D', N)


def none1():
    [p for p in t]


yl = [random.randrange(N) for i in range(N)]


def list1():
    [yl[p] for p in t]


yd = dict((i, yl[i]) for i in range(N))


def dict1():
    [yd[p] for p in t]


yo = ODict((i, yl[i]) for i in range(N))


def odict1():
    [yo[p] for p in t]


yod = yo.data


def od_data1():
    [yod[p] for p in t]


tdata = [yl]
# , [random.randrange(N) for i in range(N)]]
tab = TableDataset(data=tdata)
tab_map = tab.list


def Tab1():
    tab.getRow(t)


def Tab_mp1():
    [tuple(c[p] for c in tab_map) for p in t]


idx = Indexed(indexPattern=[0])
idx.data = [yl]  # value look-up needs to confine the value set
idx.updateToc()


def Ind1():
    [idx.vLookUp(p, return_index=0) for p in yl]


def Ind_m1():
    idx.vLookUp(yl, return_index=0, multiple=True)


it = IndexedTableDataset(data=tdata)
it.indexPattern = [0]
it.data = [yl]
it.updateToc()


def IndTab1():
    [it.vLookUp(p, return_index=0) for p in yl]


def IndT_m1():
    it.vLookUp(yl, return_index=0, multiple=True)


a = list1()
assert dict1() == a
assert odict1() == a
assert od_data1() == a
assert Tab1() == a
assert Tab_mp1() == a
assert Ind1() == a
assert Ind_m1() == a
assert IndTab1() == a
assert IndT_m1() == a

res = cmpthese(loop,
               [none1, list1, dict1, odict1, od_data1,
                Tab1, Tab_mp1, Ind1, Ind_m1,
                IndTab1, IndT_m1],
               repeat=rpt)

print(pprint_cmp(res))

del yl, yd, yo, yod, t, tdata, tab, it
#############

gc.collect()
gc.disable()

m = 1000
n = 1024
N2 = m*n
print('\n2D', N2, m, n)

t2 = [(random.randrange(m), random.randrange(n)) for i in range(N2)]


def none2():
    [1 for p, q in t2]


yl2 = [[random.randrange(N) for i in range(n)] for j in range(m)]


def list2():
    [yl2[p][q] for p, q in t2]


yd2 = dict(((j, i), t2[j*m+i]) for i in range(n) for j in range(m))


def dict2():
    [yd2[p] for p in t2]


yo2 = ODict(yd2)


def odict2():
    [yo2[p] for p in t2]


tdata2 = list(zip(*((j, i, t2[j*m+i]) for i in range(n) for j in range(m))))
idx2 = Indexed(indexPattern=[0, 1])
idx2.data = tdata2
idx2.updateToc()


def Ind2():
    [idx2.vLookUp(p, return_index=0) for p in t2]


def Ind_m2():
    idx2.vLookUp(t2, return_index=0, multiple=1)


it2 = IndexedTableDataset(data=tdata2)
it2.indexPattern = [0, 1]
it2.updateToc()


def IndTab2():
    [it2.vLookUp(p, return_index=0) for p in t2]


def IndT_m2():
    it2.vLookUp(t2, return_index=0, multiple=True)


res = cmpthese(1,
               [none2, list2, dict2, odict2,
                Ind2, Ind_m2, IndTab2, IndT_m2],
               repeat=rpt)

print(pprint_cmp(res))

del t2, yl2, yd2, yo2, tdata2, it2
gc.collect()
gc.disable()

LoL = [[c1, c2]
       for c1, c2 in zip(string.ascii_lowercase, string.ascii_uppercase)] * (N//25)


def lc_d(item='d'):
    return item in [i for sub in LoL for i in sub]


def ge_d(item='d'):
    return item in (y for x in LoL for y in x)


def any_lc_d(item='d'):
    return any(item in x for x in LoL)


def any_gc_d(item='d'):
    return any([item in x for x in LoL])


def lc_z(item='z'):
    return item in [i for sub in LoL for i in sub]


def ge_z(item='z'):
    return item in (y for x in LoL for y in x)


def any_lc_z(item='z'):
    return any(item in x for x in LoL)


def any_gc_z(item='z'):
    return any([item in x for x in LoL])


res = cmpthese(loop,
               [lc_d, ge_d, any_gc_d, any_gc_z, any_lc_d, any_lc_z, lc_z, ge_z],
               repeat=rpt)

print(pprint_cmp(res))

exit()

# https://stackoverflow.com/a/11922913/13472124
seq_of_tups = (('a', 1), ('b', 2), ('c', 3))


def d124a(): G


'a' in map(itemgetter(0), seq_of_tups)


def d124b():
    'a' in (x[0] for x in seq_of_tups)


def d124c():
    any(x == 'a' for x, y in seq_of_tups)


def d124d():
    any(x[0] == 'a' for x in seq_of_tups)


print(pprint_cmp(cmpthese(10000, [d124a, d124b, d124c, d124d], repeat=3)))



# -*- coding: utf-8 -*-

from ..dataset.classes import Classes
from ..utils.getconfig import getConfig

from .pnsconfig import pnsconfig as pc
import datetime
import time
import sys
import pwd
import os
from os import listdir, chown, chmod, environ, setuid, setgid
from pathlib import Path
import types
from subprocess import Popen, PIPE, TimeoutExpired, run as srun
from flask import Flask, jsonify, abort, make_response, request, url_for
from flask_httpauth import HTTPBasicAuth
import filelock

# from .logdict import logdict
# '/var/log/pns-server.log'
# logdict['handlers']['file']['filename'] = '/tmp/server.log'


def setuplogging():
    import logging.config
    import logging
    from . import logdict

    # create logger
    logging.config.dictConfig(logdict.logdict)
    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)
    if sys.version_info[0] > 2:
        logging.getLogger("urllib3").setLevel(logging.WARN)
    return logging


logging = setuplogging()
logger = logging.getLogger(__name__)

pc.update(getConfig())
logger.setLevel(pc['logginglevel'])
logger.debug('logging level %d' % (logger.getEffectiveLevel()))


def getUidGid(username):
    """ returns the UID and GID of the named user.
    """

    try:
        uid = pwd.getpwnam(username).pw_uid
    except KeyError as e:
        msg = 'Cannot get UserID for ' + username + \
            '. check config. ' + str(e) + trbk(e)
        logger.error(msg)
        uid = -1
    # do if platform supports.
    try:
        gid = pwd.getpwnam(username).pw_gid
    except KeyError as e:
        msg = 'Cannot get GroupID for ' + username + \
            '. check config. ' + str(e) + trbk(e)
        gid = -1
        logger.error(msg)

    return uid, gid


app = Flask(__name__)
auth = HTTPBasicAuth()


@app.before_first_request
def init_skeleton_module():
    global pc
    # effective group of current process
    uid, gid = getUidGid(pc['serveruser'])
    # logger.info
    print("Set process to %s's uid %d and gid %d..." %
          (pc['serveruser'], uid, gid))
    os.setuid(uid)
    os.setgid(gid)

    ptsuid, ptsgid = getUidGid(pc['ptsuser'])
    if gid not in os.getgrouplist(pc['ptsuser'], ptsgid):
        logger.error('ptsuser %s must be in the group of serveruser %s.' %
                     (pc['ptsuser'], pc['serveruser']))
        sys.exit(2)

    # setup user class mapping
    clp = pc['userclasses']
    logger.debug('User class file '+clp)
    if clp == '':
        Classes.updateMapping()
    else:
        clpp, clpf = os.path.split(clp)
        sys.path.insert(0, os.path.abspath(clpp))
        # print(sys.path)
        pcs = __import__(clpf.rsplit('.py', 1)[
            0], globals(), locals(), ['PC'], 0)
        pcs.PC.updateMapping()
        Classes.updateMapping(pcs.PC.mapping)
        logger.debug('User classes: %d found.' % len(pcs.PC.mapping))

# logger.debug('logging file %s' % (logdict['handlers']['file']['filename']))


def setOwnerMode(p, username):
    """ makes UID and GID set to those of serveruser given in the config file. This function is usually done by the initPTS script.
    """

    logger.debug('set owner, group to %s, mode to 0o775' % username)

    uid, gid = getUidGid(username)
    if uid == -1 or gid == -1:
        return None
    try:
        chown(str(p), uid, gid)
        chmod(str(p), mode=0o775)
    except Exception as e:
        msg = 'cannot set input/output dirs owner to ' + \
            username + ' or mode. check config. ' + str(e) + trbk(e)
        logger.error(msg)
        return None

    return username


def checkpath(path):
    """ Checks  the directories and creats if missing.

    for data exchange between pns server and  pns PTS.
    path: str. can be resolved with Path.
    """
    logger.debug(path)
    p = Path(path).resolve()
    un = pc['serveruser']
    if p.exists():
        if not p.is_dir():
            msg = str(p) + ' is not a directory.'
            logger.error(msg)
            return None
        else:
            # if path exists and can be set owner and group
            if p.owner() != un or p.group() != un:
                msg = str(p) + ' owner %s group %s. Should be %s.' % \
                    (p.owner(), p.group(), un)
                logger.warning(msg)
    else:
        # path does not exist

        msg = str(p) + ' does not exist. Creating...'
        logger.debug(msg)
        p.mkdir(mode=0o775, parents=True, exist_ok=True)
        logger.info(str(p) + ' directory has been made.')

    #logger.info('Setting owner, group, and mode...')
    if not setOwnerMode(p, un):
        return None

    logger.debug('checked path at ' + str(p))
    return p


# import requests
# from http.client import HTTPConnection
# HTTPConnection.debuglevel = 1

# @auth.verify_password
# def verify(username, password):
#     """This function is called to check if a username /
#     password combination is valid.
#     """
#     if not (username and password):
#         return False
#     return username == pc['node']['username'] and password == pc['node']['password']
@auth.verify_password
def verify_password(username, password):
    print(username + "/" + password)
    if not (username and password):
        return False
    elif username == pc['auth_user'] and password == pc['auth_pass']:
        return True
    else:
        return False
    # else:
    #     password = str2md5(password)
    #     try:
    #         conn = mysql.connector.connect(host = pc['mysql']['host'], port=pc['mysql']['port'], user =pc['mysql']['user'], password = pc['mysql']['password'], database = pc['mysql']['database'])
    #         if conn.is_connected():
    #             logger.info("connect to db successfully")
    #             cursor = conn.cursor()
    #             cursor.execute("SELECT * FROM userinfo WHERE userName = '" + username + "' AND password = '" + password + "';" )
    #             record = cursor.fetchall()
    #             if len(record) != 1:
    #                 logger.info("User : " + username + " auth failed")
    #                 conn.close()
    #                 return False
    #             else:
    #                 conn.close()
    #                 return True
    #         else:
    #             return False
    #     except Error as e:
    #         logger.error("Connect to database failed: " +str(e))


APIs = {}


def makepublicAPI(ops):
    """ Provides API specification for command given. """
    api = []
    o = APIs[ops]
    for cmd in o['cmds'].keys():
        cs = o['cmds'][cmd]
        if not issubclass(cs.__class__, tuple):  # e.g. 'run':run
            c = cs
            kwds = {}
        else:  # e.g. 'sleep': (dosleep, dict(ops='1'))
            c = cs[0]
            kwds = cs[1]
        desc = c.__doc__ if isinstance(c, types.FunctionType) else c
        d = {}
        d['description'] = desc
        d['URL'] = url_for(o['func'],
                           cmd=cmd,
                           **kwds,
                           _external=True)
        api.append(d)
    # print('******* ' + str(api))
    return api


print(id(APIs))


@app.route(pc['baseurl'] + '/', methods=['GET'])
@app.route(pc['baseurl'] + '/api', methods=['GET'])
def get_apis():
    """ Makes a page for APIs described in module variable APIs. """

    logger.debug('APIs %s' % (APIs.keys()))
    ts = time.time()
    l = [(a, makepublicAPI(a)) for a in APIs.keys()]
    w = {'APIs': dict(l), 'timestamp': ts}
    logger.debug('ret %s' % (str(w)[:100] + ' ...'))
    return jsonify(w)


@app.errorhandler(400)
def bad_request(error):
    ts = time.time()
    w = {'error': 'Bad request.', 'message': str(error), 'timestamp': ts}
    return make_response(jsonify(w), 400)


@app.errorhandler(401)
def unauthorized(error):
    ts = time.time()
    w = {'error': 'Unauthorized. Authentication needed to modify.',
         'message': str(error), 'timestamp': ts}
    return make_response(jsonify(w), 401)


@app.errorhandler(404)
def not_found(error):
    ts = time.time()
    w = {'error': 'Not found.', 'message': str(error), 'timestamp': ts}
    return make_response(jsonify(w), 404)


@app.errorhandler(409)
def conflict(error):
    ts = time.time()
    w = {'error': 'Conflict. Updating.',
         'message': str(error), 'timestamp': ts}
    return make_response(jsonify(w), 409)


logger.debug('END OF '+__file__)



# -*- coding: utf-8 -*-

from ..utils.common import mstr
from .listener import DatasetListener
from .datawrapper import DataWrapperMapper
from .composite import Composite
from .annotatable import Annotatable
from .attributable import Attributable

import pdb

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class AbstractComposite(Attributable, Annotatable, Composite, DataWrapperMapper, DatasetListener):
    """ an annotatable and attributable subclass of Composite. 
    """

    def __init__(self, **kwds):
        # pdb.set_trace()
        super(AbstractComposite, self).__init__(**kwds)

    def __repr__(self, **kwds):
        ''' toString(level=1, **kwds)
        '''
        return self.toString(level=1, **kwds)

    def toString(self, level=0, matprint=None, trans=True, beforedata='', **kwds):
        """ matprint: an external matrix print function
        trans: print 2D matrix transposed. default is True.
        """
        cn = self.__class__.__name__
        s = '# ' + cn + '\n' +\
            mstr(self.__getstate__(), level=level, **kwds)
        d = cn + '-datasets =\n'
        d += self._sets.toString(level=level,
                                 matprint=matprint, trans=trans, **kwds)
        return '\n\n'.join((x for x in (s, beforedata, d) if len(x)))



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Quantifiable(object):
    """ A Quantifiable object is a numeric object that has a unit.
    $ x.unit = ELECTRON_VOLTS
    $ print x.unit
    eV [1.60218E-19 J]"""

    def __init__(self, unit=None, typecode=None, **kwds):
        """ Has a unit and a typecode (as in array.array.typecodes).

        """
        self.setUnit(unit)
        self.setTypecode(typecode)
        #print(__name__ + str(kwds))
        super(Quantifiable, self).__init__(**kwds)

    @property
    def unit(self):
        return self.getUnit()

    @unit.setter
    def unit(self, unit):
        self.setUnit(unit)

    def getUnit(self):
        """ Returns the unit related to this object."""
        return self._unit

    def setUnit(self, unit):
        """ Sets the unit of this object. """
        self._unit = unit

    @property
    def typecode(self):
        return self.getTypecode()

    @typecode.setter
    def typecode(self, typecode):
        self.setTypecode(typecode)

    def getTypecode(self):
        """ Returns the typecode related to this object."""
        return self._typecode

    def setTypecode(self, typecode):
        """ Sets the typecode of this object. """
        self._typecode = typecode



# -*- coding: utf-8 -*-
import pdb
# from .odict import ODict
import logging
import json
import copy
import codecs
from collections.abc import Collection, Mapping
import sys
if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = (str, bytes)
else:
    PY3 = False
    strset = (str, unicode)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class SerializableEncoderAll(json.JSONEncoder):
    """ can encode parameter and product etc such that they can be recovered
    with deserialize().
    Python 3 treats string and unicode as unicode, encoded with utf-8,
    byte blocks as bytes, encoded with utf-8.
    Python 2 treats string as str and unicode as unicode, encoded with utf-8,
    byte blocks as str, encoded with utf-8
    """

    def default(self, obj):
        # logger.debug
        # print('&&&& %s %s' % (str(obj.__class__), str(obj)))
        if PY3 and issubclass(obj.__class__, bytes):
            return dict(code=codecs.encode(obj, 'hex'), _STID='bytes')
        if not PY3 and issubclass(obj.__class__, str):
            return dict(code=codec.encode(obj, 'hex'), _STID='bytes')
        if obj is Ellipsis:
            return {'obj': '...', '_STID': 'ellipsis'}
        # print(obj.__getstate__())

        if issubclass(obj.__class__, Serializable):
            return obj.__getstate__()
        print('%%%' + str(obj.__class__))
        return

        # Let the base class default method raise the TypeError
        d = json.JSONEncoder.default(self, obj)
        print('encoded d=' + d)
        return d

    # https://stackoverflow.com/a/63455796/13472124
    base = (str, int, float, bool, type(None))

    def _preprocess(self, obj):
        """ this all only work on the first level of nested objects """
        oc = obj.__class__
        ocn = type(obj).__name__

        #print('%%%*****prepro ' + ocn)
        # pdb.set_trace()
        # if issubclass(oc, self.base):
        #     # mainly to process string which is a collections (bellow)
        #     return obj
        # elif 0 and issubclass(oc, (Serializable, bytes)):
        #     if issubclass(oc, dict):
        #         # if is both __Getstate__ and Mapping, insert _STID, to a copy
        #         o = copy.copy(obj)
        #         o['_STID'] = obj._STID
        #         return o
        #     return obj
        # elif isinstance(obj, list):
        #     return obj
        # elif issubclass(oc, (Mapping)):
        #     # if all((issubclass(k.__class__, self.base) for k in obj)):
        #     if True:
        #         # JSONEncoder can handle the keys
        #         if isinstance(obj, dict):
        #             return obj
        #         else:
        #             return {'obj': dict(obj), '_STID': ocn}
        #     else:
        #         # This handles the top-level dict keys
        #         return {'obj': [(k, v) for k, v in obj.items()], '_STID': ocn}
        if issubclass(oc, (Collection)):
            return {'obj': list(obj), '_STID': ocn}
        # elif obj is Ellipsis:
        #     return {'obj': '...', '_STID': ocn}

        else:
            return obj

    def iterencode(self, obj, **kwds):
        return super().iterencode(self._preprocess(obj), **kwds)


class SerializableEncoder(json.JSONEncoder):
    """ can encode parameter and product etc such that they can be recovered
    with deserialize().
    Python 3 treats string and unicode as unicode, encoded with utf-8,
    byte blocks as bytes, encoded with utf-8.
    Python 2 treats string as str and unicode as unicode, encoded with utf-8,
    byte blocks as str, encoded with utf-8
    """

    def default(self, obj):
        try:
            # print('%%%' + str(obj.__class__))
            # Let the base class default method raise the TypeError
            d = json.JSONEncoder.default(self, obj)
            # print('d=' + d)
        except TypeError as err:
            try:
                # logger.debug
                # print('&&&& %s %s' % (str(obj.__class__), str(obj)))
                if PY3 and issubclass(obj.__class__, bytes):
                    return dict(code=codecs.encode(obj, 'hex'), _STID='bytes')
                if not PY3 and issubclass(obj.__class__, str):
                    return dict(code=codec.encode(obj, 'hex'), _STID='bytes')
                if obj is Ellipsis:
                    return {'obj': '...', '_STID': 'ellipsis'}
                # print(obj.serializable())
                return obj.serializable()
            except Exception as e:
                print('Serialization failed.' + str(e))
                raise


#    obj = json.loads(jstring)

def serialize(o, cls=None, **kwds):
    """ return JSON using special encoder SerializableEncoder """
    if not cls:
        cls = SerializableEncoder
    return json.dumps(o, cls=cls, **kwds)


class Serializable(object):
    """ mh: Can be serialized.
    Has a _STID  instance property to show its class information.
    """

    def __init__(self, **kwds):
        super(Serializable, self).__init__(**kwds)
        sc = self.__class__
        #print('@@@ ' + sc.__name__, str(issubclass(sc, dict)))
        if 0 and issubclass(sc, dict):
            self['_STID'] = sc.__name__
        else:
            self._STID = sc.__name__

    def serialized(self, indent=None):
        return serialize(self, indent=indent)

    def __getstate__(self):
        """ returns an odict that has all state info of this object.
        Subclasses should override this function.
        """
        raise NotImplementedError()

    def __setstate__(self, state):
        for name in self.__getstate__().keys():
            self.__setattr__(name, state[name])

    def __reduce_ex__(self, protocol):
        def func(): return self.__class__()
        args = tuple()
        state = self.__getstate__()
        return func, args, state

    def __reduce__(self):
        return self.__reduce_ex__(4)

    def serializable(self):
        """ Can be encoded with serializableEncoder """

        return self.__getstate__()



# -*- coding: utf-8 -*-
from .serializable import Serializable
from .odict import ODict
from .eq import DeepEqual
from .copyable import Copyable
#from .metadata import ParameterTypes
import datetime
import pdb
from collections import OrderedDict

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


# A UTC class.


class UTC(datetime.tzinfo):
    """UTC
    https://docs.python.org/2.7/library/datetime.html?highlight=datetime#datetime.tzinfo
    """

    ZERO = datetime.timedelta(0)
    HOUR = datetime.timedelta(hours=1)

    def utcoffset(self, dt):
        return UTC.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return UTC.ZERO


utcobj = datetime.timezone.utc


class FineTime(Copyable, DeepEqual, Serializable):
    """ Atomic time (SI seconds) elapsed since the TAI epoch
    of 1 January 1958 UT2. The resolution is one microsecond and the
    allowable range is: epoch + /-290, 000 years approximately.

    This has the following advantages, compared with the standard class:

    It has better resolution(microseconds)
    Time differences are correct across leap seconds
    It is immutable
    """
    """ Te starting date in UTC """
    EPOCH = datetime.datetime(1958, 1, 1, 0, 0, 0, tzinfo=utcobj)

    """ number of TAI units in a second """
    RESOLUTION = 1000000  # microsecond

    """ """
    DEFAULT_FORMAT = '%Y-%m-%dT%H:%M:%S.%f UTC'  # ISO

    RETURNFMT = '%s.%06d'

    """ """
    TIMESPEC = 'microseconds'

    def __init__(self, date=None, format=None, **kwds):
        """ Initiate with a UTC date or an integer TAI"""
        leapsec = 0  # leap seconds to be implemented

        self.format = self.DEFAULT_FORMAT if format is None else format
        self.setTime(date)
        # logger.debug('date= %s TAI = %d' % (str(date), self.tai))
        super(FineTime, self).__init__(**kwds)

    @property
    def time(self):
        return self.getTime()

    @time.setter
    def time(self, time):
        self.setTime(time)

    def getTime(self):
        """ Returns the time related to this object."""
        return self.tai

    def setTime(self, time):
        """ Sets the time of this object. 

        If an integer is given, it will be taken as the TAI.
        If a datetime object or a string code is given, the timezone will be set to UTC.
        """

        if not issubclass(time.__class__, int):
            pass  # pdb.set_trace()
        if time is None:
            self.tai = 0
        elif issubclass(time.__class__, int):
            self.tai = time
        elif issubclass(time.__class__, datetime.datetime):
            if time.tzinfo is None:
                d = time.replace(tzinfo=utcobj)
            else:
                d = time
            self.tai = self.datetimeToFineTime(d)
        elif issubclass(time.__class__, str):
            # try:
            #     # TODO: xxx
            #     self.tai = int(time)
            # except ValueError:
            d = datetime.datetime.strptime(time, self.format)
            d1 = d.replace(tzinfo=datetime.timezone.utc)
            self.tai = self.datetimeToFineTime(d1)
        else:
            raise TypeError('%s must be an integer, a datetime object, or a string, but its type is %s.' % (
                str(time), type(time).__name__))

    def getFormat(self):
        return self.format

    def microsecondsSinceEPOCH(self):
        """ Return the rounded integer number of microseconds since the epoch: 1 Jan 1958. """
        return int(self.tai * self.RESOLUTION / FineTime.RESOLUTION+0.5)

    def subtract(self, time):
        """ Subract the specified time and return the difference
        in microseconds. """
        return self.tai - time.tai

    def datetimeToFineTime(self, dtm):
        """ Return given  Python Datetime as a FineTime to the precision of the input. Rounded to the last digit. Unit is decided by RESOLUTION."""
        leapsec = 0  # leap seconds to be implemented
        return int(self.RESOLUTION *
                   ((dtm - self.EPOCH).total_seconds() + leapsec)+0.5)

    def toDatetime(self, tai):
        """ Return given FineTime as a Python Datetime. """
        leapsec = 0  # leap seconds to be implemented
        return datetime.timedelta(seconds=(float(tai) / self.RESOLUTION - leapsec))\
            + self.EPOCH

    def toDate(self):
        """ Return this time as a Python Datetime. """

        return self.toDatetime(self.tai)

    def isoutc(self, format='%Y-%m-%dT%H:%M:%S'):
        """ Returns a String representation of this objet in ISO format without timezone. sub-second set to TIMESPEC.

        format: time format. default '%Y-%m-%dT%H:%M:%S' prints like 2019-02-17T12:43:04.577000 """
        s = self.tai / self.RESOLUTION
        sub = s - int(s)
        dt = datetime.timedelta(seconds=(s)) + self.EPOCH
        # return dt.isoformat(timespec=self.TIMESPEC)
        return self.RETURNFMT % (dt.strftime(format), int(sub*self.RESOLUTION+0.5))

    def toString(self, level=0, width=0, **kwds):
        """ Returns a String representation of this object according to self.format.
        prints like 2019-02-17T12:43:04.577000 TAI(...)"""
        tais = str(self.tai) if hasattr(self, 'tai') else 'Unknown_TAI'
        if level == 0:
            if width:
                tstr = self.isoutc(
                    format='%Y-%m-%d\n%H:%M:%S') + '\n%s' % tais
                s = tstr
            else:
                tstr = self.isoutc() + ' TAI(%s)' % tais
                s = self.__class__.__name__ + \
                    '{' + tstr + ' fmt=' + self.format + '}'
        elif level == 1:
            if width:
                tstr = self.isoutc(
                    format='%Y-%m-%d\n%H:%M:%S') + '\n%s' % tais
            else:
                tstr = self.isoutc() + ' TAI(%s)' % tais
            s = tstr
        else:
            s = tais
        return s

    def __lt__(self, obj):
        """ can compare TAI directly """
        if 1:
            # if type(obj).__name__ in ParameterTypes.values():
            return self.tai < obj
        else:
            return super(FineTime, self).__lt__(obj)

    def __gt__(self, obj):
        """ can compare TAI directly """
        if 1:
            # if type(obj).__name__ in ParameterTypes.values():
            return self.tai > obj
        else:
            return super(FineTime, self).__gt__(obj)

    def __le__(self, obj):
        """ can compare TAI directly """
        if 1:
            # if type(obj).__name__ in ParameterTypes.values():
            return self.tai <= obj
        else:
            return super(FineTime, self).__le__(obj)

    def __ge__(self, obj):
        """ can compare TAI directly """
        if 1:
            # if type(obj).__name__ in ParameterTypes.values():
            return self.tai >= obj
        else:
            return super(FineTime, self).__ge__(obj)

    def __repr__(self):
        return self.toString()

    def __str__(self):
        return self.toString()

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(tai=self.tai,
                           format=self.format,
                           _STID=self._STID)


class FineTime1(FineTime):
    """ Same as FineTime but Epoch is 2017-1-1 0 UTC and unit is millisecond"""
    EPOCH = datetime.datetime(2017, 1, 1, 0, 0, 0, tzinfo=utcobj)
    RESOLUTION = 1000  # millisecond
    RETURNFMT = '%s.%03d'
    TIMESPEC = 'milliseconds'



# -*- coding: utf-8 -*-

from . import metadata

import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MetaDataHolder(object):
    """ Object holding meta data. 
    mh: object for compatibility with python2
    """

    def __init__(self, meta=None, **kwds):

        if meta is None:
            meta = metadata.MetaData()
        self.setMeta(meta)
        super(MetaDataHolder, self).__init__(**kwds)

    def getMeta(self):
        """ Returns the current MetaData container of this object. 
        Cannot become a python property because setMeta is in Attributable
        """
        return self._meta

    def hasMeta(self):
        """ whether the metadata holder is present.
        During initialization subclass of MetaDataHolder may need to know if the metadata holder has been put in place with is method.
        """

        return hasattr(self, '_meta')



# -*- coding: utf-8 -*-


from ..utils.common import lls, ld2tk
from .serializable import Serializable

import pprint
from functools import lru_cache
from collections import OrderedDict

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class CircularCallError(RuntimeError):
    pass


def deepcmp(obj1, obj2, seenlist=None, verbose=False, eqcmp=False):
    """ Recursively descends into obj1's every member, which may be
    set, list, dict, ordereddict, (or ordereddict subclasses) and
    any objects with '__class__' attribute,
    compares every member found with its counterpart in obj2.
    Returns None if finds no difference, a string of explanation
    otherwise.
    Detects cyclic references.

    eqcmp: if True, use __eq__ or __cmp__ if the objs have them. If False only use as the last resort. default True.
    """
    # seen and level are to be used as nonlocal variables in run()
    # to overcome python2's lack of nonlocal type this method is usded
    # https://stackoverflow.com/a/28433571
    class _context:
        if seenlist is None:
            seen = []
        else:
            seen = seenlist
        level = 0

    def run(o1, o2, v=False, eqcmp=True, default=None):
        #
        # nonlocal seen
        # nonlocal level
        id1, id2 = id(o1), id(o2)
        if id1 == id2:
            if v:
                print('they are the same object.')
            return None
        pair = {id1, id2}
        c = o1.__class__
        c2 = o2.__class__
        if v:
            _context.level += 1
            print('deepcmp level %d seenlist length %d' %
                  (_context.level, len(_context.seen)))
            print('1 ' + str(c) + lls(o1, 75))
            print('2 ' + str(c2) + lls(o2, 75))
        if pair in _context.seen:
            msg = 'deja vue %s' % str(pair)
            raise CircularCallError(msg)
        _context.seen.append(pair)
        if c != c2:
            if v:
                print('type diff')
            return ' due to diff types: ' + c.__name__ + ' and ' + c2.__name__
        dc, sc, fc, tc, lc = dict, set, frozenset, tuple, list

        has_eqcmp = (hasattr(o1, '__eq__') or hasattr(
            o1, '__cmp__')) and not issubclass(c, DeepEqual)
        if eqcmp and has_eqcmp:
            if v:
                print('obj1 has __eq__ or __cmp__ and not using deepcmp')
            # checked in-seen to ensure whst follows will not cause RecursionError
            try:
                t = o1 == o2
            except CircularCallError as e:
                if v:
                    print('Get circular call using eq/cmp: '+str(e))
                pass
            else:
                if t:
                    return None
                else:  # o1 != o2:
                    s = ' due to "%s" != "%s"' % (lls(o1, 155), lls(o2, 155))
                    return s
        try:
            # this is not good if len() is delegated
            # if hasattr(o1, '__len__') and len(o1) != len(o2):
            if hasattr(o1, '__len__') and len(o1) != len(o2):
                return ' due to diff %s lengths: %d and %d (%s, %s)' %\
                    (c.__name__, len(o1), len(o2), str(list(o1)), str(list(o2)))
        except AttributeError:
            pass
        if issubclass(c, dc):
            if v:
                print('Find dict or subclass')
                print('check keys')
            if issubclass(c, OrderedDict):
                #
                r = run(list(o1.keys()), list(o2.keys()), v=v, eqcmp=eqcmp)
            else:
                #  dict
                r = run(set(o1.keys()), set(o2.keys()), v=v, eqcmp=eqcmp)
            if r is not None:
                return " due to diff " + c.__name__ + " keys" + r
            if v:
                print('check values')
            for k in o1.keys():
                if k not in o2:
                    return ' due to o2 has no key=%s' % (lls(k, 155))
                r = run(o1[k], o2[k], v=v, eqcmp=eqcmp)
                if r is not None:
                    s = ' due to diff values for key=%s' % (lls(k, 155))
                    return s + r
            return None
        elif issubclass(c, (sc, fc, tc, lc)):
            if v:
                print('Find set, tuple, or list.')
            if issubclass(c, (tc, lc)):
                if v:
                    print('Check tuple or list.')
                for i in range(len(o1)):
                    r = run(o1[i], o2[i], v=v, eqcmp=eqcmp)
                    if r is not None:
                        return ' due to diff at index=%d' % (i) + r
                return None
            else:
                if v:
                    print('Check set/frozenset.')
                if 1:
                    if o1.difference(o2):
                        return ' due to at leasr one in the foremer not in the latter'
                    else:
                        return None
                else:
                    oc = o2.copy()
                    for m in o1:
                        found = False
                        for n in oc:
                            r = run(m, n, v=v, eqcmp=eqcmp)
                            if r is None:
                                found = True
                                break
                        if not found:
                            return ' due to %s not in the latter' % (lls(m, 155))
                        oc.remove(n)
                    return None
        elif hasattr(o1, '__dict__'):
            if v:
                print('obj1 has __dict__')
            r = run(o1.__dict__, o2.__dict__, v=v, eqcmp=eqcmp)
            if r:
                return ' due to o1.__dict__ != o2.__dict__' + r
            else:
                return None
        elif hasattr(o1, '__iter__') and hasattr(o1, '__next__') or \
                hasattr(o1, '__getitem__'):
            # two iterators are equal if all comparable properties are equal.
            return None
        elif has_eqcmp:
            # last resort
            if o1 == o2:
                return None
            else:
                return ' according to __eq__ or __cmp__'
        else:  # o1 != o2:
            if v:
                print('no way')
            s = ' due to no reason found for "%s" == "%s"' % (
                lls(o1, 155), lls(o2, 155))
    return run(obj1, obj2, v=verbose, eqcmp=eqcmp)


class StateEqual(object):
    """ Equality tested by hashed state.
    """

    def __hash__(self):

        @lru_cache(maxsize=16)
        def cached_hash(t):
            return hash(t)

        t = tuple(self.__getstate__().values())
        try:
            return cached_hash(t)
        except TypeError:
            """ lists/dicts recursively changed to tuples/frozensets before hashed """
            t = ld2tk(t)
            return cached_hash(t)

    def equals(self, obj, verbose=False, **kwds):
        return self.__eq__(obj, verbose=verbose)

    def __eq__(self, obj, verbose=False, **kwds):
        """ compares hash. """
        try:
            h1, h2 = self.__hash__(), obj.__hash__()
        except AttributeError:
            return False
        if verbose:
            print('hashes ', h1, h2)
        return h1 == h2

    def a__ne__(self, obj):
        return not self.__eq__(obj)


class DeepcmpEqual(object):
    """ mh: Can compare key-val pairs of another object
    with self. False if compare with None
    or exceptions raised, e.g. obj does not have items()
    """

    def equals(self, obj, verbose=False):
        r = self.diff(obj, [], verbose=verbose)
        # logging.debug(r)
        return r is None

    def __eq__(self, obj):
        return self.equals(obj)

    def __ne__(self, obj):
        return not self.__eq__(obj)

    def diff(self, obj, seenlist, verbose=False):
        """ recursively compare components of list and dict.
        until meeting equality.
        seenlist: a list of classes that has been seen. will not descend in to them.
        """
        if issubclass(self.__class__, Serializable):
            if issubclass(obj.__class__, Serializable):
                r = deepcmp(self.__getstate__(),
                            obj.__getstate__(), seenlist=seenlist, verbose=verbose)
            else:
                return('different classes')
        else:
            r = deepcmp(self, obj, seenlist=seenlist, verbose=verbose)
        return r


class EqualDict(object):
    """ mh: Can compare key-val pairs of another object
    with self. False if compare with None
    or exceptions raised, e.g. obj does not have items()
    """

    def equals(self, obj, verbose=False):

        if obj is None:
            return False
        try:
            if self.__dict__ != obj.__dict__:
                if verbose:
                    print('@@ diff \n' + lls(self.__dict__) +
                          '\n>>diff \n' + lls(obj.__dict__))
                return False
        except Exception as err:
            # print('Exception in dict eq comparison ' + lls(err))
            return False
        return True

    def __eq__(self, obj):
        return self.equals(obj)

    def __ne__(self, obj):
        return not self.__eq__(obj)


class EqualODict(object):
    """ mh: Can compare order and key-val pairs of another object
    with self. False if compare with None
    or exceptions raised, e.g. obj does not have items()
    """

    def equals(self, obj, verbose=False):
        if obj is None:
            return False
        try:
            return list(self.items()) == list(obj.items())
        except Exception:
            return False
        return True

    def __eq__(self, obj):
        return self.equals(obj)

    def __ne__(self, obj):
        return not self.__eq__(obj)


DeepEqual = StateEqual



import sys
import traceback
import pdb


def lumberjack():
    bright_side_of_death()


def bright_side_of_death():
    return tuple()[0]


try:
    lumberjack()
except IndexError:
    exc_type, exc_value, exc_traceback = sys.exc_info()
    print("*** print_tb:")
    traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
    print("*** print_exception:")
    # exc_type below is ignored on 3.5 and later
    traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    print("*** print_exc:")
    traceback.print_exc(limit=2, file=sys.stdout)
    print("*** format_exc, first and last line:")
    formatted_lines = traceback.format_exc().splitlines()
    print(formatted_lines[0])
    print(formatted_lines[-1])
    print("*** format_exception:")
    # exc_type below is ignored on 3.5 and later
    print(repr(traceback.format_exception(exc_type, exc_value,
                                          exc_traceback)))
    print("*** extract_tb:")
    print(repr(traceback.extract_tb(exc_traceback)))
    print("*** format_tb:")
    print(repr(traceback.format_tb(exc_traceback)))
    print("*** tb_lineno:", exc_traceback.tb_lineno)
    pdb.set_trace()



# -*- coding: utf-8 -*-

from ..utils.common import lls
from ..dataset.deserialize import deserialize
from ..dataset.serializable import serialize
from ..pal.productstorage import ProductStorage
from ..pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from ..pal.query import MetaQuery, AbstractQuery
from ..pal.urn import makeUrn, parseUrn
from ..dataset.product import Product
from ..utils.common import fullname, trbk

# from .db_utils import check_and_create_fdi_record_table, save_action

# import mysql.connector
# from mysql.connector import Error

import sys
import os
import json
import time
import pprint
from flask import request, make_response

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    # strset = (str, unicode)
    strset = str
    from urlparse import urlparse

# from .logdict import logdict
# '/var/log/pns-server.log'
# logdict['handlers']['file']['filename'] = '/tmp/server.log'

from .server_skeleton import logging, checkpath, pc, Classes, app, auth, APIs

logger = logging.getLogger(__name__)


# =============HTTP POOL=========================

# the httppool that is local to the server
schm = 'server'
basepath = PoolManager.PlacePaths[schm]
poolpath = os.path.join(basepath, pc['api_version'])


@app.before_first_request
def init_httppool_server():
    """ Init a global HTTP POOL """
    global PM

    logger.setLevel(pc['logginglevel'])
    logging.getLogger("filelock").setLevel(logging.INFO)
    logger.debug('logging level %d' % (logger.getEffectiveLevel()))

    PM = PoolManager
    if PM.isLoaded(DEFAULT_MEM_POOL):
        logger.debug('cleanup DEFAULT_MEM_POOL')
        PM.getPool(DEFAULT_MEM_POOL).removeAll()
    logger.debug('Done cleanup PoolManager.')
    logger.debug('ProcID %d 1st reg %s' % (os.getpid(),
                                           str(app._got_first_request))
                 )
    PM.removeAll()
    if checkpath(poolpath) is None:
        logger.error('Store path %s unavailable.' % poolpath)
        sys.exit(-2)

    load_all_pools()


def load_all_pools():
    """
    Adding all pool to server pool storage.
    """
    alldirs = set()

    def getallpools(path):
        allfilelist = os.listdir(path)
        for file in allfilelist:
            filepath = os.path.join(path, file)
            if os.path.isdir(filepath):
                alldirs.add(file)
        return alldirs
    path = poolpath
    logger.debug('loading all from ' + path)

    alldirs = getallpools(path)
    for poolname in alldirs:
        poolurl = schm + '://' + os.path.join(poolpath, poolname)
        PM.getPool(poolname=poolname, poolurl=poolurl)
        logger.info("Registered pool: %s in %s" % (poolname, poolpath))


# Check database
# check_and_create_fdi_record_table()


@ app.route(pc['baseurl'])
def get_pools():
    return str(pstore.getPools())


@ app.route(pc['baseurl'] + '/sn' + '/<string:prod_type>' + '/<string:pool_id>', methods=['GET'])
def get_pool_sn(prod_type, pool_id):
    """ Return the total count for the given product type and pool_id."""
    logger.debug('### method %s prod_type %s poolID %s***' %
                 (request.method, prod_type, pool_id))
    res = 0
    path = os.path.join(poolpath, pool_id)
    if os.path.exists(path):
        for i in os.listdir(path):
            if i[-1].isnumeric() and prod_type in i:
                res = res+1
    return str(res)


@ app.route(pc['baseurl'] + '/<path:pool>', methods=['GET', 'POST', 'DELETE'])
@ auth.login_required
def httppool(pool):
    """
    APIs for CRUD products, according to path and methods and return results.

    - GET: /pool_id/hk ==> return pool_id housekeeping
                 /pool_id/product_class/index ==> return product
                 /pool_id/hk ===> return pool_id Housekeeping data; urns, classes, and tags
                 /pool_id/hk/{urns, classes, tags} ===> return pool_id urns or classes or tags

    - POST: /pool_id ==> Save product in requests.data in server

    - DELETE: /pool_id ==> Wipe all contents in pool_id
                         /pool_id/product_class/index ==> remove specified products in pool_id
    """
    username = request.authorization.username
    paths = pool.split('/')
    ts = time.time()
    logger.debug('*** method %s paths %s ***' % (request.method, paths))
    if request.method == 'GET':
        # TODO modify client loading pool , prefer use load_HKdata rather than load_single_HKdata, because this will generate enormal sql transaction
        if paths[-2] == 'hk' and paths[-1] in ['classes', 'urns', 'tags']:  # Retrieve single HKdata
            result, msg = load_single_HKdata(paths)
            # save_action(username=username, action='READ', pool=paths[0])
        elif paths[-1] == 'hk':  # Load all HKdata
            result, msg = load_HKdata(paths)
            # save_action(username=username, action='READ', pool=paths[0])
        elif paths[-1].isnumeric():  # Retrieve product
            result, msg = load_product(paths)
            # save_action(username=username, action='READ', pool=paths[0])
        else:
            result = None
            msg = 'Unknow request: ' + pool

    if request.method == 'POST' and paths[-1].isnumeric() and request.data != None:
        data = deserialize(request.data)
        if request.headers.get('tag') is not None:
            tag = request.headers.get('tag')
        else:
            tag = None
        result, msg = save_product(data, paths, tag)
        # save_action(username=username, action='SAVE', pool=paths[0])

    if request.method == 'DELETE':
        if paths[-1].isnumeric():
            result, msg = delete_product(paths)
            # save_action(username=username, action='DELETE', pool=paths[0] +  '/' + paths[-2] + ':' + paths[-1])
        else:
            result, msg = delete_pool(paths)
            # save_action(username=username, action='DELETE', pool=paths[0])

    #w = {'result': result, 'msg': msg, 'timestamp': ts}
    # make a json string
    r = '"null"' if result is None else str(result)
    w = '{"result": %s, "msg": %s, "timestamp": %f}' % (
        r, json.dumps(msg), ts)
    # logger.debug(pprint.pformat(w, depth=3, indent=4))
    s = w  # serialize(w)
    logger.debug(lls(s, 120))
    resp = make_response(s)
    resp.headers['Content-Type'] = 'application/json'
    return resp


def delete_product(paths):
    """ removes specified product from pool
    """

    typename = paths[-2]
    indexstr = paths[-1]
    poolname = '/'.join(paths[0: -2])
    poolurl = schm + '://' + os.path.join(poolpath, poolname)
    urn = makeUrn(poolname=poolname, typename=typename, index=indexstr)
    # resourcetype = fullname(data)

    if not PM.isLoaded(poolname):
        result = '"FAILED"'
        msg = 'Pool not found: ' + poolname
        return result, msg
    logger.debug('DELETE product urn: ' + urn)
    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.remove(urn)
        msg = 'remove product ' + urn + ' OK.'
    except Exception as e:
        result = '"FAILED"'
        msg = 'Unable to remove product: ' + urn + \
            ' caused by ' + str(e) + ' ' + trbk(e)
    return result, msg


def delete_pool(paths):
    """ Remove contents of a pool
    Checking if the pool exists in server, and removing or returning exception message to client.
    """

    poolname = '/'.join(paths)
    poolurl = schm + '://' + os.path.join(poolpath, poolname)
    # resourcetype = fullname(data)

    if not PM.isLoaded(poolname):
        result = '"FAILED"'
        msg = 'Pool not found: ' + poolname
        return result, msg
    logger.debug('DELETE POOL' + poolname)
    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.removeAll()
        msg = 'Wipe pool ' + poolname + ' OK.'
    except Exception as e:
        result = '"FAILED"'
        msg = 'Unable to wipe pool: ' + poolname + \
            ' caused by ' + str(e) + ' ' + trbk(e)
    return result, msg


def save_product(data, paths, tag=None):
    """Save products and returns URNs.

    Saving Products to HTTPpool will have data stored on the server side. The server only returns URN strings as a response. ProductRefs will be generated by the associated httpclient pool which is the front-end on the user side.


    Returns a URN object or a list of URN objects. 
    """

    typename = paths[-2]
    index = str(paths[-1])
    poolname = '/'.join(paths[0: -2])
    fullpoolpath = os.path.join(poolpath, poolname)
    poolurl = schm + '://' + fullpoolpath
    # resourcetype = fullname(data)

    if checkpath(fullpoolpath) is None:
        result = '"FAILED"'
        msg = 'Pool directory error: ' + fullpoolpath
        return result, msg

    logger.debug('SAVE product to: ' + poolurl)
    logger.debug(str(id(PM._GlobalPoolList)) + ' ' + str(PM._GlobalPoolList))

    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.saveProduct(
            product=data, tag=tag, geturnobjs=True, serialized=True)
        msg = 'Save data to ' + poolurl + ' OK.'
    except Exception as e:
        result = '"FAILED"'
        msg = 'Exception : ' + str(e) + ' ' + trbk(e)
    return result, msg


def load_product(paths):
    """Load product
    """

    typename = paths[-2]
    indexstr = paths[-1]
    poolname = '/'.join(paths[0: -2])
    poolurl = schm + '://' + os.path.join(poolpath, poolname)
    urn = makeUrn(poolname=poolname, typename=typename, index=indexstr)
    # resourcetype = fullname(data)

    if not PM.isLoaded(poolname):
        result = '"FAILED"'
        msg = 'Pool not found: ' + poolname
        return result, msg

    logger.debug('LOAD product: ' + urn)
    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.loadProduct(urn=urn, serialized=True)
        msg = ''
    except Exception as e:
        result = '"FAILED"'
        msg = 'Exception : ' + str(e) + ' ' + trbk(e)
    return result, msg


def load_HKdata(paths):
    """Load HKdata of a pool
    """

    hkname = paths[-1]
    poolname = '/'.join(paths[0: -1])
    poolurl = schm + '://' + os.path.join(poolpath, poolname)
    # resourcetype = fullname(data)

    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        c, t, u = poolobj.readHK(serialized=True)
        # make a json string
        result = '{"classes": %s, "tags": %s, "urns": %s}' % (c, t, u)
        msg = ''
    except Exception as e:
        result = '"FAILED"'
        msg = 'Exception : ' + str(e) + ' ' + trbk(e)
        raise e
    return result, msg


def load_single_HKdata(paths):
    """Load classes or urns or tags of a pool
    """

    hkname = paths[-1]
    # paths[-2] is 'hk'
    poolname = '/'.join(paths[: -2])
    poolurl = schm + '://' + os.path.join(poolpath, poolname)
    # resourcetype = fullname(data)

    try:
        poolobj = PM.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.readHK(hkname, serialized=True)
        msg = ''
    except Exception as e:
        result = '"FAILED"'
        msg = 'Exception : ' + str(e) + ' ' + trbk(e)
    return result, msg


@ app.route(pc['baseurl'] + '/<string:cmd>', methods=['GET'])
def getinfo(cmd):
    ''' returns init, config, run input, run output.
    '''
    logger.debug('getr %s' % (cmd))

    msg = ''
    ts = time.time()
    try:
        if cmd == 'pnsconfig':
            result, msg = pc, ''
        else:
            result, msg = -1, cmd + ' is not valid.'
    except Exception as e:
        result, msg = -1, str(e) + trbk(e)
    w = {'result': result, 'message': msg, 'timestamp': ts}

    s = serialize(w)
    logger.debug(s[:] + ' ...')
    resp = make_response(s)
    resp.headers['Content-Type'] = 'application/json'
    return resp


# API specification for this module
ModAPIs = {'GET':
           {'func': 'get_pool_sn',
            'cmds': {'sn': 'the Serial Number'}
            },
           'PUT':
           {
           },
           'POST':
           {'func': 'httppool',
            'cmds': {}
            },
           'DELETE':
           {'func':  'httppool',
               'cmds': {}
            }}


# Use ModAPIs contents for server_skeleton.get_apis()
APIs.update(ModAPIs)
logger.debug('END OF '+__file__)



# -*- coding: utf-8 -*-
from .comparable import Comparable
from ..dataset.serializable import Serializable
from ..dataset.odict import ODict
from ..dataset.eq import DeepEqual
from ..dataset.classes import Classes
from ..utils.common import fullname
import sys
import os
from collections import OrderedDict

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    strset = (str, unicode)
    from urlparse import urlparse


def makeUrn(poolname, typename, index):
    """ assembles a URN with infos of the pool, the resource type, and the index

    index: int or string
    """
    return 'urn:' + poolname + ':' + typename + ':' + str(index)


class Urn(DeepEqual, Serializable, Comparable):
    """ The object representation of the product URN string. The
    memory consumed by sets of this object are much less than sets
    of URN strings.

    Only when the class types in URN string are not in classpath,
    the urn object will consume equals or a little more than URN string
    as the object has to hold the original urn string. However this should
    be considered as exceptional cases.

    Using this object representation also help to avoid parsing cost of    URN string.
    mh:
    URN format::

       urn:``poolname``:``resourcetype``:``serialnumber``

    where

    :poolname: Also called poolID, a (optionally path-like) string. Its format should follow that of a directory name, without leading or trailing ``/``.
    :resourcetype: (fully qualified) class name of the resource (usually Product)
    :serialnumber: internal index.

    The ``poolname`` in a URN is a label. Although pool providers could use '/' in it to introduce internal heirachy. The ``PoolURL`` is used to give practical information of a pool. It is a local set-up detail that is supposed to be hidden from pool users. The format::

         ``schme``://``place````poolpath``/``poolname``

    :scheme: Implementation protocol including ``file`` for :class:`LocalPool`, ``mem`` for :class:`MemPool`, ``http``, ``https`` for :class:`HttpclientPool`.
    :place: IP:port such as``192.168.5.6:8080`` for ``http`` and ``https`` schemes, or an empty string for ``file`` and ``mem`` schemes.
    :poolpath: The full path where the pool is stored, can have API version.  :meth:`ProductPool.transformpath` is used to map it further.
    :poolname: same as in URN

    URN is immutable.
    """

    def __init__(self, urn=None, poolname=None, cls=None, index=None, poolurl=None, **kwds):
        """
        Creates the URN object with the urn string or components.

        give urn and optional poolurl, or all poolname, cls, index arguments.
        if urn is given and pool, class, etc are also specified,
        the latter are ignored. else the URN object is constructed from them. 
        Urn(u) will make a Urn object out of u.

        All arguements are None by default.
        """
        super(Urn, self).__init__(**kwds)

        if urn is None:
            if cls is None or poolname is None or index is None:
                if cls is None and poolname is None and index is None:
                    self._scheme = None
                    self._place = None
                    self._poolname = None
                    self._class = None
                    self._index = None
                    self._poolpath = None
                    self._urn = None
                    return
                else:
                    raise ValueError(
                        'give urn and optional poolurl, or all poolname, cls, index arguments')
            urn = makeUrn(poolname=poolname,
                          typename=fullname(cls),
                          index=index)
        self.setUrn(urn, poolurl=poolurl)

    @property
    def urn(self):
        """ property
        """
        return self.getUrn()

    @urn.setter
    def urn(self, urn):
        """ property
        """
        self.setUrn(urn)

    def setUrn(self, urn, poolurl=None):
        """ parse urn to get poolname, resource, index.
        """
        if hasattr(self, '_urn') and self._urn and urn:
            raise TypeError('URN is immutable.')

        poolname, resourcetype, index = parseUrn(urn)

        cls = Classes.mapping[resourcetype.split('.')[-1]]

        self._poolname = poolname
        self._class = cls
        self._index = index
        self._urn = urn

        if poolurl:
            poolpath, scheme, place, poolname = parse_poolurl(
                poolurl, poolname)
            self._poolpath = poolpath
            self._scheme = scheme
            self._place = place
        else:
            self._poolpath = None
            self._scheme = None
            self._place = None

    def getUrn(self):
        """ Returns the urn in this """
        return self._urn

    def getType(self):
        """ Returns class type of Urn
        """
        return self._class

    def getTypeName(self):
        """ Returns class type name of Urn.
        """
        return fullname(self._class)

    def getIndex(self):
        """ Returns the product index.
        """
        return self._index

    def getScheme(self):
        """ Returns the urn scheme.
        """
        return self._scheme

    def getUrnWithoutPoolId(self):
        return fullname(self._class) + ':' + str(self._index)

    @ property
    def place(self):
        return self.getPlace()

    def getPlace(self):
        """ Returns the netloc in this """
        return self._place

    def getPoolpath(self):
        """ returns the poolpath stored
        """
        return self._poolpath

    @ property
    def pool(self):
        """ returns the poolname.
        """
        return self.getPoolId()

    def getPoolId(self):
        """ Returns the pool URN in this """
        return self._poolname

    def getPool(self):
        """ Returns the pool name in this """
        return self.getPoolId()

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(urn=self._urn,
                           _STID=self._STID)

    def __repr__(self):
        return self.__class__.__name__ + ' ' + self._urn

    def toString(self):
        return self.__class__.__name__ + \
            '{ %s, scheme:%s, place:%s, pool:%s, type:%s, index:%d, poolpath: %s}' % (
                self._urn,
                self._scheme,
                self._place,
                self._poolname,
                self._class,  # .__name__,
                self._index,
                self._poolpath
            )

    def a__hash__(self):
        """ has the URN string. """
        return hash(self._urn)


"""
    Product URNs are more complicated. For example if the urn is ``urn:file://c:/tmp/mypool/proj1.product:322``
    * poolname, also called poolURN or poolID ``file://c:/tmp/mypool``,
    * resource type (usually class) name ``proj1.product``,
    * serial number in string ``'322'``,
    * scheme ``file``,
    * place ``c:`` (with ip and port if given),
    * poolpath ``c:/tmp/mypool`` , (with ip and port if given),
"""


def parseUrn(urn):
    """
    Checks the URN string is valid in its form and splits it.

    A Product URN has several segment. For example if the urn is ``urn:mypool/v2:proj1.product:322``
    * poolname, also called poolURN or poolID, optionally path-like: ``mypool/v2``,
    * resource type (usually class) name ``proj1.product``,
    * index number  ``322``,

    returns poolname, resourceclass, index. if urn is None or empty returns (None,None,None) 

    """

    if urn is None or urn == '':
        return (None, None, None)
    if not issubclass(urn.__class__, strset):
        raise ValueError('a string is needed: ' + str(urn))
    # is a urn str?
    sp1 = urn.split(':')
    if sp1[0].lower() != 'urn':
        raise ValueError('Not a URN: ' + urn)
    # this is a product URN
    if len(sp1) != 4:
        # must have 4 segments
        raise ValueError('bad urn: ' + str(sp1))

    index = int(sp1[3])
    resourcetype = sp1[2]
    poolname = sp1[1]
    if len(poolname) * len(resourcetype) == 0:
        # something for a name
        raise ValueError('empty poolname or typename in urn: ' + urn)
    return poolname, resourcetype, index


def parse_poolurl(url, poolhint=None):
    """
    Disassambles a pool URL.

    A Pool URL is in the form of a URL that preceeds its poolname part. It is generated to desribe a pool's name, its location, and its access scheme. For example:
    * file:///tmp/mydata for pool ```mydata```,
    * file:///d:/data/test2/v2 for pool ``test2/v2``
    * mem:///dummy for pool ``dummy``
    * https://10.0.0.114:5000/v0.6/obs for a httpclientpool ``obs``
    * server;///tmp/data/0.4/test for a httppool ``test`` used on a server.

    input:
    * url: to be decomposed.
    * poolhint:  A urn or a poolname (the first distinctive substring) needs to be given if the poolname has more than one level.

    returns: poolpath, scheme, place, poolname

    * scheme: ``file``, ``mem``, ``http`` ...
    * place: ``10.2.1.10:2000``, empty string for ``file`` scheme
    * poolpath: The part between ``place`` and ``poolhint``. for ``file`` or ``server`` schemes, e.g. it is``/c:/tmp`` in ``http://localhost:9000/c:/tmp/mypool/`` with ``poolhint`` keyword arguement not given, or given as ``mypool`` (or ``myp` or 'my' ...). for ``http`` and ``https`` schemes, it is e.g. ``/0.6/tmp`` in ``https://10.0.0.114:5000/v0.6/tmp/mypool`` with ``poolhint`` keyword arguement not given, or given as ``mypool`` (or ``myp` or 'my' ...), the meaning of poolpath is subject to interpretation by the  server. Note that trailing blank and ``/`` are ignored, and stripped in the output.
    poolname: poolname from poolurl string

    returns (None, None,None,None) if url is None or empty.
    """

    if url is None or url == '':
        return (None, None, None, None)
    if not issubclass(url.__class__, strset):
        raise ValueError('a string is needed: ' + str(url))

    sp1 = url.split(':')
    if len(sp1) > 3:  # after scheme and a possible windows path
        raise ValueError(
            'a pool URN can have no more than 2 \':\'.')

    pr = urlparse(url)
    scheme = pr.scheme       # file
    place = pr.netloc
    # Note that trailing blank and ``/`` are ignored.
    path = pr.path.strip().rstrip('/')
    # convenient access path
    # get the poolname
    if poolhint:
        ps = poolhint.split(':')
        poolin = ps[1] if ps[0].lower() == 'urn' else ps[0]
        pind = path.index(poolin)
        poolname = path[pind:]
        poolpath = path[:pind].rstrip('/')
    else:
        # the last level is assumed to be the poolname
        sp = path.rsplit('/', 1)
        poolname = sp[1]
        poolpath = sp[0]

    poolpath = place + poolpath if scheme in ('file') else poolpath
    return poolpath, scheme, place, poolname


class UrnUtils():

    @ staticmethod
    def checkUrn(identifier):
        """ Throw a ValueError  if the identifier is not a legal URN."""
        if not issubclass(identifier.__class__, str):
            raise ValueError('Not a string: %s' % str(identifier))
        return parseUrn(identifier)

    @ staticmethod
    def containsUrn(poolobj,  urn):
        """ Informs whether a URN belongs to the given pool. """

        return poolobj.exists(urn)

    @ staticmethod
    def extractRecordIDs(urns):
        """ Extracts product IDs (serial numbers) from a set of urns. """
        ids = []
        for u in urns:
            pn, prod, sn = parseUrn(u)
            ids.append(sn)
        return ids

    @ staticmethod
    def getClass(urn):
        """ Get the class contained in a URN. """
        pn, prod, sn = parseUrn(urn)
        return Classes.mapping[prod.rsplit('.', 1)[1]]

    @ staticmethod
    def getClassName(urn):
        """ Get the class name contained in a URN. """
        pn, prod, sn = parseUrn(urn)
        return prod

    @ staticmethod
    def getLater(urn1, urn2):
        """ Returns the later of two urns. """
        pn1, prod1, sn1 = parseUrn(urn1)
        pn2, prod2, sn2 = parseUrn(urn2)
        return urn1 if sn1 > sn2 else urn2

    @ staticmethod
    def getPool(urn,  pools):
        """ Returns the pool corresponding to the pool id inside the given urn. 

        pools: ProductPool or subclass
        """
        if issubclass(urn.__class__, Urn):
            urn = urn.urn
        pn, prod, sn = parseUrn(urn)
        for p in pools:
            if pn == p.getId():
                return p
        raise KeyError(pn + ' not found in pools')

    @ staticmethod
    def getPoolId(urn):
        """ Returns the pool id part of the URN. """
        pn, prod, sn = parseUrn(urn)
        return pn

    @ staticmethod
    def getProductId(urn):
        """ Returns the product id part of the URN, that is, the last token. """
        pn, prod, sn = parseUrn(urn)
        return sn

    @ staticmethod
    def isUrn(identifier):
        """ Informs whether the given identifier corresponds to a URN. """

        try:
            UrnUtils.checkUrn(identifier)
        except ValueError:
            return False
        return True



# -*- coding: utf-8 -*-

from .odict import ODict
from .finetime import FineTime, FineTime1, utcobj
from .dataset import CompositeDataset
from .eq import DeepEqual, deepcmp
from collections import OrderedDict
import pdb

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class History(CompositeDataset, DeepEqual):
    """ Public interface to the history dataset. Contains the
    main methods for retrieving a script and copying the history.
    """

    def __init__(self, other=None, **kwds):
        """
        mh: The copy constructor is better not be implemented. Use copy()
        instead. Remember: not only copies the datasets,
        but also changes the history ID in the metadata and
        relevant table entries to indicate that this a new
        independent product of which the history may change.
        """
        super(History, self).__init__(**kwds)

        # Name of the table which contains the history script
        self.HIST_SCRIPT = ''
        # Name of the parameter history table
        self.PARAM_HISTORY = ''
        # Name of the task history table
        self.TASK_HISTORY = ''

    def accept(self, visitor):
        """ Hook for adding functionality to meta data object
        through visitor pattern."""
        visitor.visit(self)

    def getOutputVar(self):
        """ Returns the final output variable of the history script.
        """
        return None

    def getScript(self):
        """ Creates a Jython script from the history.
        """
        return self.HIST_SCRIPT

    def getTaskHistory(self):
        """ Returns a human readable formatted history tree.
        """
        return self.TASK_HISTORY

    def saveScript(self, file):
        """ Saves the history script to a file.
        """

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           HIST_SCRIPT=self.HIST_SCRIPT,
                           PARAM_HISTORY=self.PARAM_HISTORY,
                           TASK_HISTORY=self.TASK_HISTORY,
                           meta=self.meta,
                           _sets=self._sets,
                           _STID=self._STID)



# -*- coding: utf-8 -*-

import datetime
import traceback
import copy
import sys
import os
import pdb

from fdi.dataset.annotatable import Annotatable
from fdi.dataset.copyable import Copyable
from fdi.dataset.odict import ODict
from fdi.dataset.eq import deepcmp
from fdi.dataset.classes import Classes
from fdi.dataset.deserialize import deserialize
from fdi.dataset.quantifiable import Quantifiable
from fdi.dataset.listener import EventSender, DatasetBaseListener, EventTypes, EventType, EventTypeOf
from fdi.dataset.composite import Composite
from fdi.dataset.metadata import Parameter, NumericParameter, MetaData, StringParameter, DateParameter
from fdi.dataset.datatypes import DataTypes, DataTypeNames
from fdi.dataset.attributable import Attributable
from fdi.dataset.abstractcomposite import AbstractComposite
from fdi.dataset.datawrapper import DataWrapper, DataWrapperMapper
from fdi.dataset.dataset import ArrayDataset, TableDataset, CompositeDataset, Column
from fdi.dataset.ndprint import ndprint
from fdi.dataset.datatypes import Vector, Quaternion
from fdi.dataset.finetime import FineTime, FineTime1, utcobj
from fdi.dataset.history import History
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.product import Product
from fdi.pal.urn import Urn
from fdi.utils.checkjson import checkjson
from fdi.utils.loadfiles import loadcsv
from fdi.utils import moduleloader
from fdi.utils.common import fullname
from fdi.utils.options import opt

# import __builtins__


if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

Classes.updateMapping()

if __name__ == '__main__' and __package__ == 'tests':
    # run by python -m tests.test_dataset

    from outputs import nds2, nds3, out_TableDataset, out_CompositeDataset
else:
    # run by pytest

    # This is to be able to test w/ or w/o installing the package
    # https://docs.python-guide.org/writing/structure/
    from .pycontext import fdi

    from .outputs import nds2, nds3, out_TableDataset, out_CompositeDataset

    from .logdict import logdict
    import logging
    import logging.config
    # create logger
    logging.config.dictConfig(logdict)
    logger = logging.getLogger()
    logger.debug('%s logging level %d' %
                 (__name__, logger.getEffectiveLevel()))


def checkgeneral(v):
    # can always add attributes
    t = 'random'
    v.testattr = t
    assert v.testattr == t
    try:
        m = v.notexists
    except AttributeError as e:
        assert str(e).split()[-1] == "'notexists'", traceback.print_exc()
    except:
        traceback.print_exc()
        assert false


def test_loadcsv():
    csvf = '/tmp/testloadcsv.csv'
    a = 'as if ...'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as'], '')
    assert v[1] == ('col2', ['if'], '')
    assert v[2] == ('col3', ['...'], '')

    a = ' \t\n'+a
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as'], '')
    assert v[1] == ('col2', ['if'], '')
    assert v[2] == ('col3', ['...'], '')

    # blank line skipped
    a = a + '\n1 2. 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ' ')
    assert v[0] == ('col1', ['as', 1.0], '')
    assert v[1] == ('col2', ['if', 2.0], '')
    assert v[2] == ('col3', ['...', 3000.], '')

    # first line as header
    # pdb.set_trace()
    v = loadcsv(csvf, ' ', header=1)
    assert v[0] == ('as', [1.0], '')
    assert v[1] == ('if', [2.0], '')
    assert v[2] == ('...', [3000.], '')

    # a mixed line added. delimiter changed to ','
    a = 'as, if, ...\nm, 0.2,ev\n1, 2., 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ',', header=1)
    assert v[0] == ('as', ['m', 1.0], '')
    assert v[1] == ('if', ['0.2', 2.0], '')
    assert v[2] == ('...', ['ev', 3000.], '')

    # anothrt line added. two header lines requested -- second line taken as unit line
    a = 'as, if, ...\n A, B, R \n m, 0.2,ev\n1, 2., 3e3'
    with open(csvf, 'w') as f:
        f.write(a)
    v = loadcsv(csvf, ',', header=2)
    assert v[0] == ('as', ['m', 1.0], 'A')
    assert v[1] == ('if', ['0.2', 2.0], 'B')
    assert v[2] == ('...', ['ev', 3000.], 'R')


def test_moduleloader():

    moduleloader.main(ipath=os.path.abspath('tests'))


def test_fullname():
    assert fullname(Urn()) == 'fdi.pal.urn.Urn'
    assert fullname(Urn) == 'fdi.pal.urn.Urn'
    assert fullname('l') == 'str'


def test_opt():
    options = [
        {'long': 'helpme', 'char': 'h', 'default': False,
         'description': 'print help'},
        {'long': 'name=', 'char': 'n', 'default': 'Boo',
         'description': 'name of ship'},
        {'long': 'verbose', 'char': 'v', 'default': True,
         'description': 'print info'}
    ]
    # no args. defaults returned
    out = opt(options, [])
    assert out[0]['result'] == False
    assert out[1]['result'] == 'Boo'
    assert out[2]['result'] == True

    assert options[1]['long'] == 'name='

    # options given in short format
    out = opt(options, ['exe', '-h', '-n Awk', '-v'])
    assert out[0]['result'] == True
    # leading and trailing white spaces in args are removed
    assert out[1]['result'] == 'Awk'
    # the switch always results in True!
    assert out[2]['result'] == True

    # options given in long format
    out = opt(options, ['exe', '--helpme', '--name=Awk', '--verbose'])
    assert out[0]['result'] == True
    assert out[1]['result'] == 'Awk'
    # the switch always results in True!
    assert out[2]['result'] == True

    # type of result is determines by that of the default
    options[0]['default'] = 0
    out = opt(options, ['exe', '--helpme', '--name=Awk', '--verbose'])
    assert out[0]['result'] == 1

    # unplanned option and '--help' get exception and exits
    try:
        out = opt(options, ['exe', '--helpme', '--name=Awk', '-y'])
    except SystemExit:
        pass
    else:
        assert 0, 'failed to exit.'

    try:
        h = copy.copy(options)
        h[0]['long'] = 'help'
        out = opt(h, ['exe', '--help', '--name=Awk', '-v'])
    except SystemExit:
        pass
    else:
        assert 0, 'failed to exit.'



# -*- coding: utf-8 -*-

from ..dataset.serializable import Serializable
from ..dataset.product import Product
import pdb
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class StorageQuery(Serializable):
    """ Query on a ProductStorage. """

    def __init__(self, **kwds):
        self.query_all = True
        super(StorageQuery, self).__init__(**kwds)

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern."""
        visitor.visit(self)

    def getType(self):
        """ Get the class used in the query. """
        return self._type

    def getVariable(self):
        """ Get the variable name used in the query expression, eg "p". """
        return self._variable

    def getWhere(self):
        """ Get the query expression to be evaluated. """
        return self._where

    def retrieveAllVersions(self):
        """ Are all versions to be retrieved, or just the latest? """
        return self._allVersions


class AbstractQuery(StorageQuery):
    """ provides default implementations for the pal storage query. """

    def __init__(self, product=Product, variable='p', where='', allVersions=False, **kwds):
        """ creates an AbstractQuery with product variable name, query string or function."""
        self._type = product
        self._where = where
        self._variable = variable
        self._allVersions = allVersions

        super(AbstractQuery, self).__init__(**kwds)

    def hashCode(self):
        return None

    def toString(self):
        """
        """

        s = self.__class__.__name__ +\
            '"%s", product = %s, variable = %s, allVersions = %s' % \
            (str(self._where), str(self._type), str(
                self._variable), str(self._allVersions))
        return s


class MetaQuery(AbstractQuery):
    """ Meta data query formulates a query on the meta data of a Product.

    Typically this type of query is faster than a full query on the Product Access Layer.

    """

    def __init__(self, product=Product, where='', allVersions=False, **kwds):
        """ creates an MetaQuery with a query string or function.

        'where' is a query string or function that returns True or False.
        In the query string variable name is 'm' for a MetaData type, as in ``m = product.meta``.
        """
        # pdb.set_trace()
        super(MetaQuery, self).__init__(product=product, variable='m',
                                        where=where, allVersions=allVersions, **kwds)



# -*- coding: utf-8 -*-

from ..dataset.classes import Classes
from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize
from ..dataset.eq import deepcmp
import json
import sys
from pprint import pprint
import pdb


def getyaml():
    if 0:
        import inspect
        import collections
        import ruamel
        from ruamel.yaml import YAML
        from ruamel.yaml.representer import RoundTripRepresenter

        ruamel.yaml.add_representer(
            Classes.mapping['ODict'], RoundTripRepresenter.represent_dict)

        ruamel.yaml.add_representer(collections.OrderedDict,
                                    RoundTripRepresenter.represent_dict)
        yaml = YAML(typ='rt')
        yaml.default_flow_style = False
        from ..dataset.classes import Classes
        for n, c in Classes.mapping.items():
            if inspect.isclass(c):
                print(n+' %s' % type(c).__name__)
                yaml.register_class(c)
    else:
        from .ydump import ydump, yinit
        yinit()


def checkjson(obj, dbg=0, **kwds):
    """ seriaizes the given object and deserialize. check equality.
    """

    # dbg = True if issubclass(obj.__class__, BaseProduct) else False

    indent = 4 if dbg > 1 else None

    # if hasattr(obj, 'serialized'):
    #     js = obj.serialized(indent=indent)
    # else:
    #     js = json.dumps(obj, indent=indent)
    js = serialize(obj)
    if dbg:
        getyaml()
        try:
            if 0:
                yaml.dump(obj, sys.stdout)
            else:
                print(ydump(obj))
        except:
            print('ydump of obj failed')

        print('******** checkjsom ' + obj.__class__.__name__ +
              ' serialized: ******\n')
        print(js)
        print('*************')

    des = deserialize(js, lgb=Classes.mapping, debug=dbg)
    if dbg:
        if 0 and hasattr(des, 'meta'):
            print('des.mets ' + str((des.meta.listeners)))
        print('****** checkjson deserialized ' + str(des.__class__) +
              '******\n')
        try:
            if 0:
                yaml.dump(des, sys.stdout)
            else:
                print(ydump(des))
        except:
            print('ydump of deserialized obj failed')
            pprint(des)

        # js2 = json.dumps(des, cls=SerializableEncoder)
        # pprint('**** des     serialized: *****')
        # pprint(js)

        r = deepcmp(obj, des, **kwds)
        print('******** deepcmp ********')
        print('identical' if r is None else r)
        # print(' DIR \n' + str(dir(obj)) + '\n' + str(dir(des)))
    if 0 and issubclass(obj.__class__, BaseProduct):
        print(str(id(obj)) + ' ' + obj.toString())
        print(str(id(des)) + ' ' + des.toString())
        # obj.meta.listeners = []
        # des.meta.listeners = []

    assert obj == des, deepcmp(obj, des, **kwds)
    return des



# -*- coding: utf-8 -*-
# This is to be able to test w/ or w/o installing the package
# https://docs.python-guide.org/writing/structure/
import os
import sys
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..')))

import fdi
#from fdi import dataset
#from fdi import pal
#from fdi import pns
#from fdi import tests
# print(dir(dataset))



# -*- coding: utf-8 -*-

# Automatically generated from fdi/dataset/resources/Product.yml. Do not edit.

from collections import OrderedDict
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.finetime import FineTime


from fdi.dataset.readonlydict import ReadOnlyDict

import copy

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Product(BaseProduct):
    """ Product class (level ALL) schema 1.4 inheriting ['BaseProduct'].

Automatically generated from fdi/dataset/resources/Product.yml on 2021-02-18 22:19:05.909785.

Description:
Project level product

    Generally a Product (inheriting BaseProduct) has project-wide attributes and can be extended to define a plethora of specialized products.
    """


    def __init__(self,
                 description = 'UNKNOWN',
                 typ_ = 'Product',
                 creator = 'UNKNOWN',
                 creationDate = FineTime(0),
                 rootCause = 'UNKNOWN',
                 version = '0.8',
                 FORMATV = '1.4.0.8',
                 startDate = FineTime(0),
                 endDate = FineTime(0),
                 instrument = 'UNKNOWN',
                 modelName = 'UNKNOWN',
                 mission = '_AGS',
                 **kwds):
        """ Initializes instances with more metadata as attributes, set to default values.

        Put description keyword argument here to allow e.g. BaseProduct("foo") and description='foo'
        """
        if 'metasToBeInstalled' in kwds:
            # This class is being called probably from super() in a subclass
            metasToBeInstalled = kwds['metasToBeInstalled']
            del kwds['metasToBeInstalled']

            # must be the first line to initiate meta and get description
            super().__init__(
                metasToBeInstalled=metasToBeInstalled, **kwds)
            return
        # this class is being called directly

        # list of local variables.
        metasToBeInstalled = copy.copy(locals())
        for x in ('self', '__class__', 'kwds'):
            metasToBeInstalled.pop(x)

        global ProductInfo
        self.zInfo = ProductInfo

        #print('@1 zInfo', id(self.zInfo['metadata']), id(self), id(self.zInfo),
        #      self.zInfo['metadata']['version'], list(metasToBeInstalled.keys()))

        # must be the first line to initiate meta and get description
        super().__init__(
            metasToBeInstalled=metasToBeInstalled, **kwds)

        super().installMetas(mtbi=metasToBeInstalled, prodInfo=ProductInfo)
        #print(self.meta.keySet(), id(self.meta))


    @property
    def type(self): pass

    @type.setter
    def type(self, p): pass


    @property
    def startDate(self): pass

    @startDate.setter
    def startDate(self, p): pass


    @property
    def endDate(self): pass

    @endDate.setter
    def endDate(self, p): pass


    @property
    def instrument(self): pass

    @instrument.setter
    def instrument(self, p): pass


    @property
    def modelName(self): pass

    @modelName.setter
    def modelName(self, p): pass


    @property
    def mission(self): pass

    @mission.setter
    def mission(self, p): pass


    @property
    def version(self): pass

    @version.setter
    def version(self, p): pass


    @property
    def FORMATV(self): pass

    @FORMATV.setter
    def FORMATV(self, p): pass





_Model_Spec = {
    'name': 'Product',
    'description': 'Project level product',
    'parents': [
        'BaseProduct',
        ],
    'level': 'ALL',
    'schema': '1.4',
    'metadata': {
        'description': {
                'id_zh_cn': '描述',
                'data_type': 'string',
                'description': 'Description of this product',
                'description_zh_cn': '对本产品的描述。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'type': {
                'id_zh_cn': '产品类型',
                'data_type': 'string',
                'description': 'Product Type identification. Name of class or CARD.',
                'description_zh_cn': '产品类型。完整Python类名或卡片名。',
                'default': 'Product',
                'valid': '',
                'valid_zh_cn': '',
                'typecode': 'B',
                },
        'creator': {
                'id_zh_cn': '本产品生成者',
                'data_type': 'string',
                'description': 'Generator of this product.',
                'description_zh_cn': '本产品生成方的标识，例如可以是单位、组织、姓名、软件、或特别算法等。',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'creationDate': {
                'id_zh_cn': '产品生成时间',
                'fits_keyword': 'DATE',
                'data_type': 'finetime',
                'description': 'Creation date of this product',
                'description_zh_cn': '本产品生成时间',
                'default': 0,
                'valid': '',
                'typecode': None,
                },
        'rootCause': {
                'id_zh_cn': '数据来源',
                'data_type': 'string',
                'description': 'Reason of this run of pipeline.',
                'description_zh_cn': '数据来源（此例来自鉴定件热真空罐）',
                'default': 'UNKNOWN',
                'valid': '',
                'typecode': 'B',
                },
        'version': {
                'id_zh_cn': '版本',
                'data_type': 'string',
                'description': 'Version of product',
                'description_zh_cn': '产品版本',
                'default': '0.8',
                'valid': '',
                'typecode': 'B',
                },
        'FORMATV': {
                'id_zh_cn': '格式版本',
                'data_type': 'string',
                'description': 'Version of product schema and revision',
                'description_zh_cn': '产品格式版本',
                'default': '1.4.0.8',
                'valid': '',
                'typecode': 'B',
                },
        'startDate': {
                'id_zh_cn': '产品的标称起始时间',
                'fits_keyword': 'DATE-OBS',
                'data_type': 'finetime',
                'description': 'Nominal start time  of this product.',
                'description_zh_cn': '产品标称的起始时间',
                'default': 0,
                'valid': '',
                'valid_zh_cn': '',
                'typecode': None,
                },
        'endDate': {
                'id_zh_cn': '产品的标称结束时间',
                'fits_keyword': 'DATE-END',
                'data_type': 'finetime',
                'description': 'Nominal end time  of this product.',
                'description_zh_cn': '产品标称的结束时间',
                'default': 0,
                'valid': '',
                'valid_zh_cn': '',
                'typecode': None,
                },
        'instrument': {
                'id_zh_cn': '观测仪器名称',
                'data_type': 'string',
                'description': 'Instrument that generated data of this product',
                'description_zh_cn': '观测仪器名称',
                'default': 'UNKNOWN',
                'valid': '',
                'valid_zh_cn': '',
                'typecode': 'B',
                },
        'modelName': {
                'id_zh_cn': '样机名称',
                'fits_keyword': 'MODEL',
                'data_type': 'string',
                'description': 'Model name of the instrument of this product',
                'description_zh_cn': '观测仪器样机名称',
                'default': 'UNKNOWN',
                'valid': '',
                'valid_zh_cn': '',
                'typecode': 'B',
                },
        'mission': {
                'id_zh_cn': '任务名称',
                'fits_keyword': 'TELESCOP',
                'data_type': 'string',
                'description': 'Name of the mission.',
                'description_zh_cn': '任务名称',
                'default': '_AGS',
                'valid': '',
                'valid_zh_cn': '',
                'typecode': 'B',
                },
        },
    'datasets': {
        },
    }

ProductInfo = ReadOnlyDict(_Model_Spec)



"""
The following demostrates important dataset and pal functionalities. It was made by running fdi/resources/example.py with command ``elpy-shell-send-group-and-step [c-c c-y c-g]`` in ``emacs``. The command is further simplified to control-<tab> with the following in ~/.init.el:


.. code-bloc:: lisp
   (add-hook 'elpy-mode-hook
	  (lambda ()
	    (local-set-key [C-tab] (quote elpy-shell-send-group-and-step))))


You can copy the code from code blocks by clicking the ``copy`` icon on the top-right, with he proompts and results removed.
"""

# Import these packages needed in the tutorial
from fdi.dataset.product import Product
from fdi.dataset.metadata import Parameter, NumericParameter, MetaData, StringParameter, DateParameter
from fdi.dataset.finetime import FineTime, FineTime1
from fdi.dataset.dataset import ArrayDataset, TableDataset, Column
from fdi.dataset.classes import Classes
from fdi.pal.context import Context, MapContext
from fdi.pal.productref import ProductRef
from fdi.pal.query import AbstractQuery, MetaQuery
from fdi.pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from fdi.pal.productstorage import ProductStorage
import getpass
import os
from datetime import datetime, timezone
import logging


# initialize the white-listed class dictionary
cmap = Classes.updateMapping()

print("""
dataset
=======

First we show how to make and use components of the data model.

This section shows how to create data containers -- datasets, metadata, and Products, how to put data into the containers, read data out, modify data, remove data, inspect data.

ArrayDataset
------------
""")


# Creation with an array of data quickly
a1 = [1, 4.4, 5.4E3, -22, 0xa2]
v = ArrayDataset(a1)
# Show it. This is the same as print(v) in a non-interactive environment.
v

# Create an ArrayDataset with built-in properties set.
v = ArrayDataset(data=a1, unit='ev', description='5 elements',
                 typ_='float', default=1.0, typecode='f')
#
# add some metadats (see more about meta data below)
v.meta['greeting'] = StringParameter('Hi there.')
v.meta['year'] = NumericParameter(2020)
v

# data access: read the 2nd array element
v[2]       # 5400

# built-in properties
v.unit

# change it
v.unit = 'm'
v.unit

# iteration
for m in v:
    print(m + 1)

# a filter example
[m**3 for m in v if m > 0 and m < 40]

# slice the ArrayDataset and only get part of its data
v[2:-1]

# set data to be a 2D array
v.data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
# slicing happens on the slowest dimension.
v[0:2]

# Run this to see a demo of the ``toString()`` function::
# make a 4-D array: a list of 2 lists of 3 lists of 4 lists of 5 elements.
s = [[[[i + j + k + l for i in range(5)] for j in range(4)]
      for k in range(3)] for l in range(2)]
v.data = s
print(v.toString())

print('''
TableDataset
------------

TableDataset is mainly a dictionary containing name-Column pairs and metadata.
Columns are basically ArrayDatasets under a different name.
''')

# Create an empty TableDataset then add columns one by one
v = TableDataset()
v['col1'] = Column(data=[1, 4.4, 5.4E3], unit='eV')
v['col2'] = Column(data=[0, 43.2, 2E3], unit='cnt')
v

# Do it with another syntax, with a list of tuples and no Column()
a1 = [('col1', [1, 4.4, 5.4E3], 'eV'),
      ('col2', [0, 43.2, 2E3], 'cnt')]
v1 = TableDataset(data=a1)
v == v1

# Make a quick tabledataset -- data are list of lists without names or units
a5 = [[1, 4.4, 5.4E3], [0, 43.2, 2E3], [True, True, False], ['A', 'BB', 'CCC']]
v5 = TableDataset(data=a5)
print(v5.toString())

# access
# get names of all columns (automatically given here)
v5.getColumnNames()

# get column by name
my_column = v5['column1']       # [1, 4.4, 5.4E3]
my_column.data

# by index
v5[0].data       # [1, 4.4, 5.4E3]

# get a list of all columns' data.
# Note the slice "v5[:]" and syntax ``in``
[c.data for c in v5[:]]   # == a5

#  indexOf by name
v5.indexOf('column1')  # == u.indexOf(my_column)

#  indexOf by column object
v5.indexOf(my_column)     # 0

# set cell value
v5['column2'][1] = 123
v5['column2'][1]    # 123

# row access bu row index -- multiple and in custom order
v5.getRow([2, 1])  # [(5400.0, 2000.0, False, 'CCC'), (4.4, 123, True, 'BB')]

# or with a slice
v5.getRow(slice(0, -1))

# unit access
v1['col1'].unit  # == 'eV'

# add, set, and replace columns and rows
# column set / get
u = TableDataset()
c1 = Column([1, 4], 'sec')
# add
u.addColumn('time', c1)
u.columnCount        # 1

# for non-existing names set is addColum.
u['money'] = Column([2, 3], 'eu')
u['money'][0]    # 2
# column increases
u.columnCount        # 2

# addRow
u.rowCount    # 2

u.addRow({'money': 4.4, 'time': 3.3})
u.rowCount    # 3

# run this to see ``toString()``
ELECTRON_VOLTS = 'eV'
SECONDS = 'sec'
t = [x * 1.0 for x in range(8)]
e = [2.5 * x + 100 for x in t]
d = [765 * x - 500 for x in t]
# creating a table dataset to hold the quantified data
x = TableDataset(description="Example table")
x["Time"] = Column(data=t, unit=SECONDS)
x["Energy"] = Column(data=e, unit=ELECTRON_VOLTS)
x["Distance"] = Column(data=d, unit='m')
# metadata is optional
x.meta['temp'] = NumericParameter(42.6, description='Ambient', unit='C')
print(x.toString())

print("""
MetaData and Parameter: Parameter
---------------------------------

FDI datasets and products not only contain data, but also their metadata -- data about the "payload" data. Metadata is a collections of parameters.

A Parameter is a variable with associated information about its description, unit, type, valid ranges, default, format code etc. Type can be numeric, string, datetime, vector.

Often a parameter shows a property. So a parameter in the metadata of a dataset or product is often called a property.
""")

# Creation
# The standard way -- with keyword arguments
v = Parameter(value=9000, description='Average age', typ_='integer')
v.description   # 'Average age'

v.value   # == 9000

v.type   # == 'integer'

# test equals.
# FDI DeepEqual integerface class recursively compares all components.
v1 = Parameter(description='Average age', value=9000, typ_='integer')
v.equals(v1)

# more readable 'equals' syntax
v == v1

# make them not equal.
v1.value = -4
v.equals(v1)   # False

# math syntax
v != v1  # True

# NumericParameter with two valid values and a valid range.
v = NumericParameter(value=9000, valid={
                     0: 'OK1', 1: 'OK2', (100, 9900): 'Go!'})
# There are thee valid conditions
v

# The current value is valid
v.isValid()


# check if other values are valid according to specification of this parameter
v.validate(600)  # valid

v.validate(20)  # invalid


print("""
MetaData and Parameter: Metadata
--------------------------------

A dict-like container for named parameters.
""")

# Creation. Start with numeric parameter.
a1 = 'weight'
a2 = NumericParameter(description='How heavey is the robot.',
                      value=60, unit='kg', typ_='float')
# make an empty MetaData instance.
v = MetaData()
# place the parameter with a name
v.set(a1, a2)
# get the parameter with the name.
v.get(a1)   # == a2

# add more parameter. Try a string type.
v.set(name='job', newParameter=StringParameter('pilot'))
# get the value of the parameter
v.get('job').value   # == 'pilot'

# access parameters in metadata
# a more readable way to set/get a parameter than "v.set(a1,a2)", "v.get(a1)"
v['job'] = StringParameter('waitress')
v['job']   # == waitress

# same result as...
v.get('job')

# Date type parameter use International Atomic Time (TAI) to keep time,
# in 1-microsecond precission
v['birthday'] = Parameter(description='was born on',
                          value=FineTime('1990-09-09T12:34:56.789098 UTC'))
# FDI use International Atomic Time (TAI) internally to record time.
# The format is the integer number of microseconds since 1958-01-01 00:00:00 UTC.
v['birthday'].value.tai


# names of all parameters
[n for n in v]   # == ['weight', 'job', 'birthday']

# remove parameter from metadata.   # function inherited from Composite class.
v.remove(a1)
v.size()  # == 2

# The value of the next parameter is valid from 0 to 31 and can be 9
valid_rule = {(0, 31): 'valid', 99: ''}
v['a'] = NumericParameter(
    3.4, 'rule name, if is "valid", "", or "default", is ommited in value string.', 'float', 2., valid=valid_rule)
v['a'].isValid()    # True

then = datetime(
    2019, 2, 19, 1, 2, 3, 456789, tzinfo=timezone.utc)
# The value of the next parameter is valid from TAI=0 to 9876543210123456
valid_rule = {(0, 9876543210123456): 'alive'}
# display typecode set to 'year' (%Y)
v['b'] = DateParameter(FineTime(then), 'date param', default=99,
                       valid=valid_rule, typecode='%Y')
# The value of the next parameter has an empty rule set and is always valid.
v['c'] = StringParameter(
    'Right', 'str parameter. but only "" is allowed.', valid={'': 'empty'}, default='cliche', typecode='B')

# The value of the next parameter is for a detector status.
# The information is packed in a byte, and if extractab;e with suitable binary masks:
# Bit7~Bit6 port status [01: port 1; 10: port 2; 11: port closed];
# Bit5 processing using the main processir or a stand-by one [0:  stand by; 1: main];
# Bit4 PPS status [0: error; 1: normal];
# Bit3~Bit0 reserved.
valid_rule = {
    (0b11000000, 0b01): 'port_1',
    (0b11000000, 0b10): 'port_2',
    (0b11000000, 0b11): 'port closed',
    (0b00100000, 0b0): 'stand_by',
    (0b00100000, 0b1): 'main',
    (0b00010000, 0b0): 'error',
    (0b00010000, 0b1): 'normal',
    (0b00001111, 0b0): 'reserved'
}
v['d'] = NumericParameter(
    0b01010110, 'valid rules described with binary masks', valid=valid_rule)
# this returns the tested value, the rule name, the heiggt and width of every mask.
v['d'].validate(0b01010110)

# string representation. This is the same as v.toString(level=0), most detailed.
print(v.toString())

# simplifed string representation, toString(level=1), also what __repr__() runs.
v

# simplest string representation, toString(level=2).
print(v.toString(level=2))

print("""
Product
-------

The data Product is at the center of FDI data model. A product has
   * zero or more datasets (say images, tables, spectra etc...). 
   * accompanying metadata,
   * history of this product: how was this data created.
""")

# Creation:
x = Product(description="product example with several datasets",
            instrument="Crystal-Ball", modelName="Mk II")
x.meta['description'].value  # == "product example with several datasets"

# The 'instrument' and 'modelName' built-in properties show the
# origin of FDI -- processing data from scientific instruments.
x.instrument  # == "Crystal-Ball"

# ways to add datasets
i0 = 6
i1 = [[1, 2, 3], [4, 5, i0], [7, 8, 9]]
i2 = 'ev'                 # unit
i3 = 'image1'     # description
image = ArrayDataset(data=i1, unit=i2, description=i3)
# put the dataset into the product
x["RawImage"] = image
# take the data out of the product
x["RawImage"].data  # == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

# Another syntax to put dataset into a product: set(name, dataset)
# Different but same function as above.
# Here no unit or description is given when making ArrayDataset
x.set('QualityImage', ArrayDataset(
    [[0.1, 0.5, 0.7], [4e3, 6e7, 8], [-2, 0, 3.1]]))
x["QualityImage"].unit  # is None

# add another tabledataset
s1 = [('col1', [1, 4.4, 5.4E3], 'eV'),
      ('col2', [0, 43.2, 2E3], 'cnt')]
x["Spectrum"] = TableDataset(data=s1)
# See the numer and types of existing datasets in the product
[type(d) for d in x.values()]

# mandatory properties are also in metadata
# test mandatory BaseProduct properties that are also metadata
a0 = "Me, myself and I"
x.creator = a0
x.creator   # == a0

# metada by the same name is also set
x.meta["creator"].value   # == a0

# change the metadata
a1 = "or else"
x.meta["creator"] = Parameter(a1)
# metada changed
x.meta["creator"].value   # == a1

# so was the property
x.creator   # == a1

# load some metadata
m = x.meta
m['ddetector'] = v['d']
# Demo ``toString()`` function.
print(x.toString())


print('''
pal
===

Products need to persist (be stored somewhere) in order to have a reference that can be used to re-create the product after its creation process ends.

Product Pool and Product References
-----------------------------------

This section shows how to store a product in a "pool" and get a reference back.
''')


# Create a product and a productStorage with a pool registered
# First disable debugging messages
logger = logging.getLogger('')
logger.setLevel(logging.WARNING)
# a pool (LocalPool) for demonstration will be create here
demopoolpath = '/tmp/demopool_' + getpass.getuser()
demopoolurl = 'file://' + demopoolpath
# clean possible data left from previous runs
os.system('rm -rf ' + demopoolpath)
if PoolManager.isLoaded(DEFAULT_MEM_POOL):
    PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
PoolManager.removeAll()

# create a prooduct and save it to a pool
x = Product(description='save me in store')
# add a tabledataset
s1 = [('energy', [1, 4.4, 5.6], 'eV'), ('freq', [0, 43.2, 2E3], 'Hz')]
x["Spectrum"] = TableDataset(data=s1)
# create a product store
pstore = ProductStorage(poolurl=demopoolurl)
# see what is in it.
pstore

# save the product and get a reference back.
prodref = pstore.save(x)
# This gives detailed information of the product being referenced
print(prodref)

# get the URN string
urn = prodref.urn
print(urn)    # urn:demopool_mh:fdi.dataset.product.Product:0

# re-create a product only using the urn
newp = ProductRef(urn).product
# the new and the old one are equal
print(newp == x)   # == True


print("""
Context
-------

A Context is a Product with References. This section shows essencial steps how product references can be stored in a context.

""")

p1 = Product(description='p1')
p2 = Product(description='p2')
# create an empty mapcontext that can carry references with name labels
map1 = MapContext(description='product with refs 1')
# A ProductRef created with the syntax of a lone product argument will use a MemPool
pref1 = ProductRef(p1)
pref1

# A productStorage with a LocalPool -- a pool on the disk.
pref2 = pstore.save(p2)
pref2.urn

# how many prodrefs do we have?
map1['refs'].size()   # == 0

# how many 'parents' do these prodrefs have before saved?
len(pref1.parents)   # == 0

len(pref2.parents)   # == 0

# add a ref to the context. Every productref has a name in a MapContext
map1['refs']['spam'] = pref1
# add the second one
map1['refs']['egg'] = pref2
# how many prodrefs do we have?
map1['refs'].size()   # == 2

# parent list of the productref object now has an entry
len(pref2.parents)   # == 1

pref2.parents[0] == map1

pref1.parents[0] == map1

# remove a ref
del map1['refs']['spam']
map1.refs.size()   # == 1

# how many prodrefs do we have?
len(pref1.parents)   # == 0

# add ref2 to another map
map2 = MapContext(description='product with refs 2')
map2.refs['also2'] = pref2
map2['refs'].size()   # == 1

# two parents
len(pref2.parents)   # == 2

pref2.parents[1] == map2

print("""
Query 
------

A ProductStorage with pools attached can be queried with tags, properties stored in metadata, or even data in the stored products, with Python syntax.
""")

# clean possible data left from previous runs
defaultpoolpath = '/tmp/pool_' + getpass.getuser()
newpoolname = 'newpool_' + getpass.getuser()
newpoolpath = '/tmp/' + newpoolname
os.system('rm -rf ' + defaultpoolpath)
os.system('rm -rf ' + newpoolpath)
if PoolManager.isLoaded(DEFAULT_MEM_POOL):
    PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
PoolManager.removeAll()
# make a productStorage
defaultpoolurl = 'file://'+defaultpoolpath
pstore = ProductStorage(poolurl=defaultpoolurl)
# make another
newpoolurl = 'file://' + newpoolpath
pstore2 = ProductStorage(poolurl=newpoolurl)

# add some products to both storages. The product properties are different.
n = 7
for i in range(n):
    # three counters for properties to be queried.
    a0, a1, a2 = 'desc %d' % i, 'fatman %d' % (i*4), 5000+i
    if i < 3:
        # Product type
        x = Product(description=a0, instrument=a1)
        x.meta['extra'] = Parameter(value=a2)
    elif i < 5:
        # a different type
        x = Context(description=a0, instrument=a1)
        x.meta['extra'] = Parameter(value=a2)
    else:
        # yet another type
        x = MapContext(description=a0, instrument=a1)
        x.meta['extra'] = Parameter(value=a2)
        x.meta['time'] = Parameter(value=FineTime1(a2))
    if i < 4:
        # some are stored in one pool
        r = pstore.save(x)
    else:
        # some the other
        r = pstore2.save(x)
    print(r.urn)
# Two pools, 7 products in 3 types
# [P P P C] [C M M]
#  0 1 2 3   4 5 6

# register the new pool above to the  1st productStorage
pstore.register(newpoolname)
len(pstore.getPools())   # == 2

# make a query on product metadata, which is the variable 'm'
# in the query expression, i.e. ``m = product.meta; ...``
# But '5000 < m["extra"]' does not work. see tests/test.py.
q = MetaQuery(Product, 'm["extra"] > 5001 and m["extra"] <= 5005')
# search all pools registered on pstore
res = pstore.select(q)
# we expect [#2, #3, #4, #5]
len(res)   # == 4

# see
[r.product.description for r in res]

# another one


def t(m):
    # query is a function
    import re
    # 'instrument' matches the regex pattern
    return re.match('.*n.1.*', m['instrument'].value)


q = MetaQuery(Product, t)
res = pstore.select(q)
# expecting [3,4]
[r.product.instrument for r in res]

# same as above but query is on the product. this is slow.
q = AbstractQuery(Product, 'p', '"n 1" in p.instrument')
res = pstore.select(q)
# [3,4]
[r.product.instrument for r in res]



# -*- coding: utf-8 -*-
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Definable(object):
    """
    for items being able to be defined with a Definition.
    """

    def __init__(self, **kwds):
        super(Definable, self).__init__(**kwds)
        self._definition = None

    def getDefinition(self):
        """
        Returns the definition associated to this definable item.
        mh: adopting http://herschel.esac.esa.int/hcss-doc-15.0/load/hcss_drm/api/index.html?herschel/ia/pal/ProductRef.html but parameterize definition does not seem worth it given the few numbers of implemented definitions
        """

        return self._definition



# -*- coding: utf-8 -*-

from itertools import count
from collections import Sequence
import sys


class Indexed():
    """ Abstract class with an index table for efficient records look-up.

    """

    def __init__(self, indexPattern=None, **kwds):
        """

        indexPattern: specifies which columns to uae to do look up.
        """
        self._indexPattern = [0] if indexPattern is None else indexPattern
        self._tableOfContent = {}
        super().__init__(**kwds)  # initialize data, meta, unit

    def getColumnsToLookup(self):
        """ returns an iterator that gives a number of sequences to looking up over.

        Defau;t is a tuple of the ``data`` attributes of every columns specified by ``indexPattern``. To be overrided for different data model.
        """

        # list of Column's arrays
        return [self.data[i] for i in self._indexPattern]

    def updateToc(self, which=None, for_records=None):
        """ Build index in format specified in indexPattern for retrieving record.
        which: an iterator that gives a number of sequences to looking up over. Default is ``getColumnsToLookup()``.
        for_records: a list or a ``Slice`` of record (row) numbers. Those are changed records that caused updating. default is all records.
        """

        # always rebuild all TODO:
        for_records = None
        self._tableOfContent = {}

        cols = self.getColumnsToLookup() if which is None else which

        ip = self._indexPattern
        one = len(ip) == 1
        if for_records is None:
            # for all records
            itr = zip(cols[0], count()) if one else zip(zip(*cols), count())
        elif issubclass(for_records.__class__, slice):
            # list of column's arrays for all slice records
            # range(sys.maxsize)[for_records] gives all valid record index numbers
            if one:
                itr = zip(cols[0][for_records],
                          range(sys.maxsize)[for_records])
            else:
                itr = zip(zip(*(c[for_records] for c in cols)),
                          range(sys.maxsize)[for_records])
        else:
            # for_records is a list
            # list of column's arrays for all listed records
            if one:
                itr = zip((cols[0][i] for i in for_records), for_records)
            else:
                itr = zip((tuple(c[i] for c in cols)
                           for i in for_records), for_records)

        self._tableOfContent.update(itr)

    @ property
    def indexPattern(self):
        return self._indexPattern

    @ indexPattern.setter
    def indexPattern(self, *key):
        """ set the key pattern used to retrieve records.

        *key: as a list of integers. taken as column numbers. future look-up will search and return the record where  a match is found in these columns. Example: a.indexPattern=[0,2] would setup to use the first and the third columns to make look-up keys. Default is the first column.
        """

        lk = len(key)
        if lk == 0:
            self._indexPattern = None
            return
        # more than on column number
        tk = []
        msg = 'Need integers or tuple of integers to specify look-up indices.'
        for k in key:
            if type(key) == int:
                tk.append(k)
            elif issubclass(key.__class__, Sequence):
                for k2 in k:
                    if type(k2) == int:
                        tk.append(k2)
                    else:
                        raise TypeError(msg)
            else:
                raise TypeError(msg)

        self._indexPattern = tk

    @ property
    def toc(self):
        """ returns  the index table of content.
        """

        return self._tableOfContent

    @ toc.setter
    def toc(self, table):
        """ sets the index table of content.

        """

        self._tableOfContent = table

    def clearToc(self):
        """ Clears the index table of content.
        """

        self._tableOfContent.clear()

    def vLookUp(self, key, return_index=True, multiple=False):
        """ Similar to Excel VLOOKUP, return all records (rows) that match the key.
        key: taken as a dictionary key unless ``multiple`` is True.
        return_index: if True (default) return index in the array of columns.
        multiple: if True (default is False) loop through key as a sequence of keys and return a sequece.
        """

        if multiple:
            if return_index:
                toc = self._tableOfContent
                return [toc[k] for k in key]
            else:
                toc = self._tableOfContent
                cols = self.data
                # return [[c[toc[k]] for c in cols] for k in key]
                return list(zip(*([c[toc[k]] for k in key] for c in cols)))
        else:
            if return_index:
                return self._tableOfContent[key]
            else:
                rec_ind = self._tableOfContent[key]
                return [c[rec_ind] for c in self.data]

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(
            indexPattern=self._indexPattern,
            toc=self._tableOfContent
        )



# -*- coding: utf-8 -*-
from ruamel.yaml import YAML
# import yaml
from collections import OrderedDict
import os
import sys
from string import Template
import pkg_resources
from datetime import datetime
import importlib

# from ..pal.context import MapContext
from ..utils.options import opt
from ..utils.common import pathjoin, trbk
from ..utils.ydump import ydump
from ..utils.moduleloader import SelectiveMetaFinder, installSelectiveMetaFinder

# a dictionary that translates metadata 'type' field to classname
from .datatypes import DataTypes, DataTypeNames

import logging

# create logger
logger = logging.getLogger(__file__)
logging.basicConfig(stream=sys.stdout,
                    format='%(asctime)s - %(levelname)7s'
                           ' - [%(filename)s:%(lineno)3s'
                           ' - %(funcName)10s()] - %(message)s',
                    datefmt="%Y-%m-%d %H:%M:%S")
logging.getLogger().setLevel(logging.DEBUG)


# make simple demo for fdi
demo = 0
# if demo is true, only output this subset.
onlyInclude = ['default', 'description',
               'data_type', 'unit', 'valid', 'fits_keyword']
# These property names are used by  fdi
reserved = ['history', 'meta', 'refs']
# only these attributes in meta
attrs = ['startDate', 'endDate', 'instrument', 'modelName', 'mission', 'type']
indent = '    '
# extra indent
ei = ''
indents = [ei + indent * i for i in range(10)]

fmtstr = {
    'integer': '{:d}',
    'short': '{:d}',
    'hex': '0x{:02X}',
    'byte': '{:d}',
    'binary': '0b{:0b}',
    'float': '{:g}',
    'string': '"{:s}"',
    'finetime': '{:d}'
}


def sq(s):
    """ add quote mark to string, depending on if ' or " in the string.
    """

    if "'" in s or '\n' in s:
        qm = '"""' if '"' in s or '\n' in s else '"'
    else:
        qm = "'"
    return '%s%s%s' % (qm, s, qm)


def getPython(val, indents, demo, onlyInclude):
    """ make productInfo and init__() code strings from given data.
    """
    infostr = ''

    if issubclass(val.__class__, dict):
        infostr += '{\n'
        code = {}
        for k, v in val.items():
            sk = str(k)
            infostr += '%s%s: ' % (indents[0], sq(sk))
            if issubclass(v.__class__, dict) and 'valid' in v:
                # v is a dict of parameter attributes
                istr, d_code = params(v, indents[1:], demo, onlyInclude)
            else:
                # headers such as name, parents, level ...
                istr, d_code = getPython(v, indents[1:], demo, onlyInclude)
            infostr += istr
            code[sk] = d_code
        infostr += indents[0] + '},\n'
    elif issubclass(val.__class__, list):
        infostr += '[\n'
        code = []
        for v in val:
            infostr += indents[0]
            if issubclass(v.__class__, dict) and 'data_type' in v:
                # val is a list of column (and 'data' in x )
                istr, d_code = params(v, indents[1:], demo, onlyInclude)
            else:
                istr, d_code = getPython(v, indents[1:], demo, onlyInclude)
            infostr += istr
            code.append(d_code)
        infostr += indents[0] + '],\n'
    else:
        pval = sq(val) if issubclass(val.__class__, (str, bytes)) else str(val)
        infostr += pval + ',\n'
        code = pval
    return infostr, code


def makeinitcode(dt, pval):
    """ python instanciation source code.

    will be like "default: FineTime1(0)"
    """
    if dt not in ['string', 'integer', 'hex', 'binary', 'float']:
        # custom classes
        t = DataTypes[dt]
        code = '%s(%s)' % (t, pval)
    elif dt in ['integer', 'hex', 'float', 'binary']:
        code = pval
    elif pval == 'None':
        code = 'None'
    else:
        code = sq(pval)
    return code


def params(val, indents, demo, onlyInclude):
    """ generates python strng for val, a parameter with a set of attribute

    see getPython
    """
    infostr = '{\n'
    code = None
    # data_type
    dt = val['data_type'].strip()
    # loop through the properties
    for pname, pv in val.items():
        # pname is like 'data_type', 'default'
        # pv is like 'string', 'foo, bar, and baz', '2', '(0, 0, 0,)'
        if demo and pname not in onlyInclude:
            continue
        if pname.startswith('valid'):
            if pv is None:
                pv = ''

            if issubclass(pv.__class__, (str, bytes)):
                s = sq(pv.strip())
            else:
                lst = []
                for k, v in pv.items():
                    if issubclass(k.__class__, tuple):
                        fs = fmtstr[dt]
                        # (,999) in yaml is ('',999) but None from inhrited class
                        foo = [fs.format(x) if x != '' and x is not None else 'None'
                               for x in k]
                        sk = '(' + ', '.join(foo) + ')'
                    else:
                        sk = fmtstr[dt].format(k)
                    lst += '\n' + '%s%s: %s,' % (indents[2], sk, sq(str(v)))
                kvs = ''.join(lst)
                if len(kvs) > 0:
                    kvs += '\n' + indents[2]
                s = '{' + kvs + '}'
        else:
            iss = issubclass(pv.__class__, (str))

            # get string representation
            pval = str(pv).strip() if iss else str(pv)
            if pname == 'default':
                code = makeinitcode(dt, pval)
            if pname in ['example', 'default']:
                # here data_type instead of input type determines the output type
                iss = val['data_type'] == 'string'
            s = sq(pval) if iss else pval
        infostr += indents[1] + '%s: %s,\n' % (sq(pname), s)
    infostr += indents[1] + '},\n'

    return infostr, code


def getCls(clp, rerun=True, exclude=None, verbose=False):
    if clp is None or len(clp.strip()) == 0:
        return {}
    if exclude is None:
        exclude = []
    if '/' not in clp and '\\' not in clp and not clp.endswith('.py'):
        print('Importing project classes from module '+clp)
        # classes path not given on command line
        pc = importlib.import_module(clp)
        pc.PC.updateMapping(rerun=rerun, exclude=exclude,
                            verbose=verbose, ignore_error=True)
        ret = pc.PC.mapping
        print(
            'Imported project classes from svom.products.projectclasses module.')

    else:
        clpp, clpf = os.path.split(clp)
        sys.path.insert(0, os.path.abspath(clpp))
        # print(sys.path)
        print('Importing project classes from file '+clp)
        pc = importlib.import_module(clpf.replace('.py', ''))
        sys.path.pop(0)
        pc.PC.updateMapping(rerun=rerun, exclude=exclude)
        ret = pc.PC.mapping
    return ret


def readyaml(ypath, ver=None):
    """ read YAML files in ypath.

    output: nm is  stem of file name. desc is descriptor, key being yaml[name]
    """
    yaml = YAML()
    desc = OrderedDict()
    fins = {}
    for findir in os.listdir(ypath):
        fin = os.path.join(ypath, findir)

        ''' The  input file name ends with '.yaml' or '.yml' (case insensitive).
        the stem name of output file is input file name stripped of the extension.
        '''
        # make it all lower case
        finl = findir.lower()
        if finl.endswith('.yml') or finl.endswith('.yaml'):
            nm = os.path.splitext(findir)[0]
        else:
            continue
        fins[nm] = fin

        # read YAML
        print('----- Reading ' + fin + '-----')
        with open(fin, 'r', encoding='utf-8') as f:
            # pyYAML d = OrderedDict(yaml.load(f, Loader=yaml.FullLoader))
            d = OrderedDict(yaml.load(f))
        if 'metadata' not in d or d['metadata'] is None:
            d['metadata'] = {}
        if 'datasets' not in d or d['datasets'] is None:
            d['datasets'] = {}
        if float(d['schema']) >= 1.0:
            pass
        if float(d['schema']) > 0.6:
            attrs = OrderedDict(d['metadata'])
            datasets = OrderedDict()
            # move primary level table to datasets
            if 'TABLE' in attrs:
                datasets['TABLE'] = {}
                datasets['TABLE']['TABLE'] = attrs['TABLE']
                del attrs['TABLE']
            if 'datasets' in d:
                datasets.update(d['datasets'])
            print('Pre-emble:\n%s' %
                  (''.join([k + '=' + str(v) + '\n'
                            for k, v in d.items() if k not in ['metadata', 'datasets']])))

            logger.debug('Find attributes:\n%s' %
                         ''.join(('%20s' % (k+'=' + str(v['default'])
                                            if 'default' in v else 'url' + ', ')
                                  for k, v in attrs.items()
                                  )))
            if float(d['schema']) > 1.1:
                itr = ('%20s' % (k+'=' + str([c for c in (v['TABLE'] if 'TABLE'
                                                          in v else [])]))
                       for k, v in datasets.items())
                logger.debug('Find datasets:\n%s' % ', '.join(itr))
            else:
                # v1.1 1.0 0.7
                itr = ('%20s' %
                       (k+'=' + str([c['name'] for c in (v['TABLE'] if 'TABLE'
                                                         in v else [])]))
                       for k, v in datasets.items())
                print('Find datasets:\n%s' % ', '.join(itr))
            desc[d['name']] = (d, attrs, datasets, fin)
        else:
            # float(d['schema']) <= 0.6:
            d2 = OrderedDict()
            metadata = OrderedDict()
            for k, v in d.items():
                if issubclass(v.__class__, dict):
                    if v['unit'] == 'None':
                        dt = v['data_type']
                        if dt in ['boolean', 'string']:
                            v['unit'] = None
                    metadata[k] = v
                else:
                    if k == 'definition':
                        d2['description'] = v
                    elif k == 'schema':
                        d2[k] = version
                    elif k == 'parent':
                        d2['parents'] = [v]
                    else:
                        d2[k] = v
            d2['metadata'] = metadata
            desc[d['name']] = d2
    return desc, fins


def output(nm, d, fins, version, verbose):

    print("Input YAML file is to be renamed to " + fins[nm]+'.old')
    fout = fins[nm]
    print("Output YAML file is "+fout)
    if 0:
        ydump(d, sys.stdout)  # yamlfile)
    else:
        os.rename(fins[nm], fins[nm]+'.old')
        with open(fout, 'w', encoding='utf-8') as yamlfile:
            ydump(d,  yamlfile)


def yamlupgrade(descriptors, fins, ypath, version, verbose):

    if float(version) > 1.0:
        for nm, daf in descriptors.items():
            d, attrs, datasets, fin = daf
            if float(d['schema']) >= float(version):
                print('No need to upgrade '+d['schema'])
                continue
            d['schema'] = version
            for pname, w in d['metadata'].items():
                dt = w['data_type']
                # no dataset yet
                if dt in ['boolean', 'string', 'finetime']:
                    del w['unit']
                if dt == 'finetime':
                    w['default'] = 0
                if 'typecode' not in w:
                    w['typecode'] = 'B' if dt == 'string' else None
                if pname == 'version':
                    v = w['default'].replace('v', '')
                    w['default'] = str(float(v) + 0.1)
            output(nm, d, fins, version, verbose)
    elif float(version) > 0.6:
        # in:v0.6 and below out:v1.0
        for nm, d in descriptors.items():
            output(nm, d, fins, version, verbose)


def dependency_sort(descriptors):
    """ sort the descriptors so that everyone's parents are to his right.

    """
    ret = []
    working_list = list(descriptors.keys())
    while len(working_list):
        # must use index to loop
        for i in range(len(working_list)):
            nm = working_list[i]
            p = descriptors[nm][0]['parents']
            nm_found_parent = False
            if len(p) == 0:
                continue
            found = set(working_list) & set(p)
            # for x in working_list:
            #    if x == nm:
            #        continue
            #    if x in p:
            if len(found):
                # parent is in working_list
                working_list.remove(nm)
                working_list.append(nm)
                nm_found_parent = True
                break
            else:
                # no one in the list is nm's superclass
                # TODO: only immediate parenthood tested
                ret.append(nm)
                working_list.remove(nm)
                break
            if nm_found_parent:
                break
        else:
            # no one in the list is free from deendency to others
            if len(working_list):
                msg = 'Cyclic dependency among ' + str(working_list)
                logger.error(msg)
                sys.exit(-5)
    return ret


def removeParent(a, b):
    """ Returns the one who is the other one's parent.
    """
    if a == b:
        logger.debug('%s and %s are the same class' % (b, a))
        return None
    tmp = "remove parent %s because it is another parent %s's"
    if issubclass(glb[a], glb[b]):
        # remove b
        logger.debug(tmp % (b, a))
        return b
    elif issubclass(glb[b], glb[a]):
        # remove a
        logger.debug(tmp % (a, b))
        return a
    else:
        return None


def noParentsParents(pn):
    removed = []
    for i in range(len(pn)-1):
        if pn[i] in removed:
            continue
        for j in range(i+1, len(pn)):
            r = removeParent(pn[i], pn[j])
            if r:
                removed.append(r)
            if r == pn[i]:
                break
    for r in removed:
        pn.remove(r)
    return pn


if __name__ == '__main__':

    print('product class generatiom')

    # schema version
    version = '1.1'

    # Get input file name etc. from command line. defaut 'Product.yml'
    cwd = os.path.abspath(os.getcwd())
    ypath = cwd
    tpath = cwd
    opath = cwd
    ops = [
        {'long': 'help', 'char': 'h', 'default': False, 'description': 'print help'},
        {'long': 'verbose', 'char': 'v', 'default': False,
         'description': 'print info'},
        {'long': 'yamldir=', 'char': 'y', 'default': ypath,
         'description': 'Input YAML file directory.'},
        {'long': 'template=', 'char': 't', 'default': tpath,
         'description': 'Product class template file directory.'},
        {'long': 'outputdir=', 'char': 'o', 'default': opath,
         'description': 'Output directory for python file.'},
        {'long': 'packagename=', 'char': 'p', 'default': '-',
         'description': 'Name of the package which the generated modules belong to.'},
        {'long': 'userclasses=', 'char': 'c', 'default': '',
         'description': 'Python file name, or a module name,  to import prjcls to update Classes with user-defined classes which YAML file refers to.'},
        {'long': 'upgrade', 'char': 'u', 'default': False,
         'description': 'Upgrade the file to current schema, to a filename + ' + version},
        {'long': 'debug', 'char': 'd', 'default': False,
         'description': 'run in pdb. type "c" to continuue.'},
    ]

    out = opt(ops)
    # print([(x['long'], x['result']) for x in out])
    verbose = out[1]['result']
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    ypath = out[2]['result']
    tpath = out[3]['result']
    opath = os.path.abspath(out[4]['result'])
    package_name = out[5]['result']
    project_class_path = out[6]['result']
    upgrade = out[7]['result']
    debug = out[8]['result']

    if debug:
        import pdb
        pdb.set_trace()

    # input file
    descriptors, fins = readyaml(ypath, version)
    if upgrade:
        yamlupgrade(descriptors, fins, ypath, version, verbose)
        sys.exit()

    if package_name == '-':
        ao = os.path.abspath(opath)
        if not ao.startswith(cwd):
            logger.error('Cannot derive package name from yaml dir and cwd.')
            exit(-3)
        package_name = ao[len(cwd):].strip('/').replace('/', '.')
    logger.debug("Package name: " + package_name)

    # Do not import modules that are to be generated. Thier source code
    # could be  invalid due to unseccessful previous runs
    importexclude = [x.lower() for x in descriptors.keys()]
    importinclude = {}

    # activate a module loader that refuses to load excluded
    installSelectiveMetaFinder()

    # include project classes for every product so that products made just
    # now can be used as parents
    from .classes import Classes
    pcl = getCls(project_class_path, rerun=True,
                 exclude=importexclude, verbose=verbose)
    glb = Classes.updateMapping(
        c=pcl, rerun=True, exclude=importexclude, verbose=verbose)
    # make a list whose members do not depend members behind
    sorted_list = dependency_sort(descriptors)
    for nm in sorted_list:
        d, attrs, datasets, fin = descriptors[nm]
        print('************** Processing ' + nm + '***********')

        # the generated source code must import these
        seen = []
        imports = 'from collections import OrderedDict\n'
        # import parent classes
        pn = d['parents']
        # remove classes that are other's parent class (MRO problem)
        if pn and len(pn):
            pn = noParentsParents(pn)

        if pn and len(pn):
            all_attrs = OrderedDict()
            for a in pn:
                if a is None:
                    continue
                modnm = glb[a].__module__
                s = 'from %s import %s\n' % (modnm, a)
                if a not in seen:
                    seen.append(a)
                    imports += s

                # get parent attributes
                mod = sys.modules[modnm]
                if hasattr(mod, '_Model_Spec'):
                    all_attrs.update(mod._Model_Spec['metadata'])
            # merge to get all attributes including parents' and self's.
            all_attrs.update(attrs)
        else:
            all_attrs = attrs

        prodname = d['name']
        # module/output file name is YAML input file "name" with lowercase
        modulename = nm.lower()

        # make output filename, lowercase modulename + .py
        fout = pathjoin(opath, modulename + '.py')
        print("Output python file is "+fout)

        # class doc
        doc = '%s class (level %s) schema %s inheriting %s.\n\nAutomatically generated from %s on %s.\n\nDescription:\n%s' % tuple(map(str, (
            prodname, d['level'], d['schema'], d['parents'],
            fin, datetime.now(), d['description'])))

        # parameter classes used in init code may need to be imported, too
        for met, val in all_attrs.items():
            a = DataTypes[val['data_type']]
            if a in glb:
                # this attribute class has module
                s = 'from %s import %s' % (glb[a].__module__, a)
                if a not in seen:
                    seen.append(a)
                    imports += s+'\n'

        # make metadata and dataset dictionaries
        d['metadata'] = all_attrs
        d['datasets'] = datasets
        infs, default_code = getPython(d, indents[1:], demo, onlyInclude)
        # remove the ',' at the end.
        infostr = (ei + '_Model_Spec = ' + infs).strip()[:-1]

        # keyword argument for __init__
        ls = []
        for x in all_attrs:
            arg = 'typ_' if x == 'type' else x
            val = default_code['metadata'][x]
            ls.append(' '*17 + '%s = %s,' % (arg, val))
        ikwds = '\n'.join(ls)
        # class properties
        pr = []
        for x in attrs:
            if x in reserved:
                logger.warning('%s is a reserved name.' % x)
            arg = x        # x + '_' if x == 'type' else x
            pr.append('    @property\n    def %s(self): pass\n\n' % (x) +
                      '    @%s.setter\n    def %s(self, p): pass\n\n' % (x, x))
        properties = '\n'.join(pr)

        # make substitution dictionary for Template
        subs = {}
        subs['WARNING'] = '# Automatically generated from %s. Do not edit.' % fin
        subs['PRODUCTNAME'] = prodname
        print('product name: %s' % subs['PRODUCTNAME'])
        subs['PARENTS'] = ', '.join(c for c in pn if c)
        print('parent class: %s' % subs['PARENTS'])
        subs['IMPORTS'] = imports
        print('import class: %s' % seen)
        subs['CLASSDOC'] = doc
        subs['PRODUCTINFO'] = infostr
        subs['INITARGS'] = ikwds
        print('productInfo=\n%s\n' % (subs['INITARGS']))
        subs['PROPERTIES'] = properties

        # subtitute the template
        if os.path.exists(os.path.join(tpath, prodname + '.template')):
            tname = os.path.join(tpath, prodname + '.template')
        elif os.path.exists(os.path.join(tpath, 'template')):
            tname = os.path.join(tpath, 'template')
        else:
            logger.error('Template file not found in %s for %s.' %
                         (tpath, prodname))
            sys.exit(-3)
        with open(tname, encoding='utf-8') as f:
            t = f.read()

        sp = Template(t).safe_substitute(subs)
        # print(sp)
        with open(fout, 'w', encoding='utf-8') as f:
            f.write(sp)
        print('Done saving ' + fout + '\n' + '='*40)

        # import the newly made module  so the following classes could use it
        importexclude.remove(modulename)
        # importlib.invalidate_caches()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)
        newp = 'fresh ' + prodname + ' from ' + modulename + \
            '.py of package ' + package_name + ' in ' + opath + '.'
        # If the last segment of package_name happens to be a module name in
        # exclude list the following import will be blocked. So lift
        # exclusion temporarily
        exclude_save = importexclude[:]
        importexclude.clear()
        try:
            _o = importlib.import_module('.' + modulename, package_name)
            glb[prodname] = getattr(_o, prodname)
        except Exception as e:
            print('Unable to import ' + newp)
            raise(e)
        importexclude.extend(exclude_save)
        print('Imported ' + newp)
        # Instantiate and dump metadata in rst format
        prod = glb[prodname]()
        fg = {'name': 15, 'value': 18, 'unit': 7, 'type': 8,
              'valid': 26, 'default': 18, 'code': 4, 'description': 25}
        sp = prod.meta.toString(tablefmt='fancy_grid', widths=fg)

        mout = pathjoin(ypath, prodname + '.txt')
        with open(mout, 'w', encoding='utf-8') as f:
            f.write(sp)
        print('Done dumping ' + mout + '\n' + '*'*40)

        if len(importexclude) == 0:
            exit(0)

        Classes.updateMapping(c=importinclude, exclude=importexclude)
        glb = Classes.mapping



from fdi.dataset.product import _Model_Spec as PPI
from .product import Product
from ..pal.context import Context, MapContext
from .finetime import FineTime

import copy


class TP(Product):
    pass


class TC(Context):
    pass


class TM(MapContext):
    pass


# sub-classing testing class
# 'version' of subclass is int, not string


class SP(Product):
    def __init__(self,
                 description='UNKNOWN',
                 typ_='SP',
                 creator='UNKNOWN',
                 version=9,
                 creationDate=FineTime(0),
                 rootCause='UNKNOWN',
                 startDate=FineTime(0),
                 endDate=FineTime(0),
                 instrument='UNKNOWN',
                 modelName='UNKNOWN',
                 mission='_AGS',
                 **kwds):
        metasToBeInstalled = copy.copy(locals())
        for x in ('self', '__class__', 'kwds'):
            metasToBeInstalled.pop(x)

        sp = {}
        sp = copy.deepcopy(PPI)
        sp['name'] = self.__class__.__name__
        sp['metadata']['version']['data_type'] = 'integer'
        sp['metadata']['version']['default'] = 9
        sp['metadata']['type']['default'] = sp['name']
        self.zInfo = sp
        assert PPI['metadata']['version']['data_type'] == 'string'
        super().__init__(metasToBeInstalled=metasToBeInstalled, **kwds)
        super().installMetas(mtbi=metasToBeInstalled, prodInfo=self.zInfo)

    @ property
    def version(self): pass

    @ version.setter
    def version(self, p): pass



from pygments.token import Keyword, Name, Comment, String, Error, \
    Number, Operator, Generic, Literal, Punctuation, Whitespace, \
    Text, STANDARD_TYPES, Token, Other
from pygments.style import Style
from pygments.styles.monokai import MonokaiStyle


class MonokaiMod(MonokaiStyle):
    styles = MonokaiStyle.styles
    styles.update({
        Comment: 'bold italic #af2',
        Literal.String.Doc: 'bold italic #af2',
        Generic.Output: 'bg:#022 ansibrightcyan',
        Text: 'bg:#022 ansibrightcyan'
    })
